from leapp.exceptions import StopActorExecutionError
from leapp.libraries import stdlib
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelInfo, KernelCmdlineArg, TargetKernelCmdlineArgTasks


def run_grubby_cmd(cmd):
    try:
        stdlib.run(cmd)
        if architecture.matches_architecture(architecture.ARCH_S390X):
            # On s390x we need to call zipl explicitly because of issue in grubby,
            # otherwise the entry is not updated in the ZIPL bootloader
            # See https://bugzilla.redhat.com/show_bug.cgi?id=1764306
            stdlib.run(['/usr/sbin/zipl'])

    except (OSError, stdlib.CalledProcessError) as e:
        raise StopActorExecutionError(
            "Failed to append extra arguments to kernel command line.",
            details={"details": str(e)})


def format_kernelarg_msgs_for_grubby_cmd(kernelarg_msgs):
    kernelarg_msgs = sorted(kernelarg_msgs, key=lambda arg: arg.key)
    kernel_args = []
    for arg in kernelarg_msgs:
        if arg.value:
            kernel_args.append('{0}={1}'.format(arg.key, arg.value))
        else:
            kernel_args.append('{0}'.format(arg.key))
    return ' '.join(kernel_args)


def modify_kernel_args_in_boot_cfg(configs_to_modify_explicitly=None):
    kernel_info = next(api.consume(InstalledTargetKernelInfo), None)
    if not kernel_info:
        return

    # Collect desired kernelopt modifications
    kernelargs_msgs_to_add = list(api.consume(KernelCmdlineArg))
    kernelargs_msgs_to_remove = []
    for target_kernel_arg_task in api.consume(TargetKernelCmdlineArgTasks):
        kernelargs_msgs_to_add.extend(target_kernel_arg_task.to_add)
        kernelargs_msgs_to_remove.extend(target_kernel_arg_task.to_remove)

    if not kernelargs_msgs_to_add and not kernelargs_msgs_to_remove:
        return  # There is no work to do

    grubby_modify_kernelargs_cmd = ['grubby', '--update-kernel={0}'.format(kernel_info.kernel_img_path)]

    if kernelargs_msgs_to_add:
        grubby_modify_kernelargs_cmd += [
            '--args', '{}'.format(format_kernelarg_msgs_for_grubby_cmd(kernelargs_msgs_to_add))
        ]

    if kernelargs_msgs_to_remove:
        grubby_modify_kernelargs_cmd += [
            '--remove-args', '{}'.format(format_kernelarg_msgs_for_grubby_cmd(kernelargs_msgs_to_remove))
        ]

    if configs_to_modify_explicitly:
        for config_to_modify in configs_to_modify_explicitly:
            cmd = grubby_modify_kernelargs_cmd + ['-c', config_to_modify]
            run_grubby_cmd(cmd)
    else:
        run_grubby_cmd(grubby_modify_kernelargs_cmd)



import os
import sys
import uuid

from leapp.cli.commands import command_utils
from leapp.cli.commands.config import get_config
from leapp.cli.commands.upgrade import breadcrumbs, util
from leapp.exceptions import CommandError, LeappError
from leapp.logger import configure_logger
from leapp.utils.audit import Execution
from leapp.utils.clicmd import command, command_opt
from leapp.utils.output import beautify_actor_exception, report_errors, report_info

# NOTE:
# If you are adding new parameters please ensure that they are set in the upgrade function invocation in `rerun`
# otherwise there might be errors.


@command('upgrade', help='Upgrade the current system to the next available major version.')
@command_opt('resume', is_flag=True, help='Continue the last execution after it was stopped (e.g. after reboot)')
@command_opt('reboot', is_flag=True, help='Automatically performs reboot when requested.')
@command_opt('whitelist-experimental', action='append', metavar='ActorName', help='Enable experimental actors')
@command_opt('debug', is_flag=True, help='Enable debug mode', inherit=False)
@command_opt('verbose', is_flag=True, help='Enable verbose logging', inherit=False)
@command_opt('no-rhsm', is_flag=True, help='Use only custom repositories and skip actions'
                                           ' with Red Hat Subscription Manager')
@command_opt('no-insights-register', is_flag=True, help='Do not register into Red Hat Insights')
@command_opt('no-rhsm-facts', is_flag=True, help='Do not store migration information using Red Hat '
                                                 'Subscription Manager. Automatically implied by --no-rhsm.')
@command_opt('enablerepo', action='append', metavar='<repoid>',
             help='Enable specified repository. Can be used multiple times.')
@command_opt('channel',
             help='Set preferred channel for the IPU target.',
             choices=['ga', 'e4s', 'eus', 'aus'],
             value_type=str.lower)  # This allows the choices to be case insensitive
@command_opt('iso', help='Use provided target RHEL installation image to perform the in-place upgrade.')
@command_opt('target', choices=command_utils.get_supported_target_versions(),
             help='Specify RHEL version to upgrade to for {} detected upgrade flavour'.format(
                 command_utils.get_upgrade_flavour()))
@command_opt('report-schema', help='Specify report schema version for leapp-report.json',
             choices=['1.0.0', '1.1.0', '1.2.0'], default=get_config().get('report', 'schema'))
@command_opt('nogpgcheck', is_flag=True, help='Disable RPM GPG checks. Same as yum/dnf --nogpgcheck option.')
@breadcrumbs.produces_breadcrumbs
def upgrade(args, breadcrumbs):
    skip_phases_until = None
    context = str(uuid.uuid4())
    cfg = get_config()
    util.handle_output_level(args)
    answerfile_path = cfg.get('report', 'answerfile')
    userchoices_path = cfg.get('report', 'userchoices')

    # Processing of parameters passed by the rerun call, these aren't actually command line arguments
    # therefore we have to assume that they aren't even in `args` as they are added only by rerun.
    only_with_tags = args.only_with_tags if 'only_with_tags' in args else None
    resume_context = args.resume_context if 'resume_context' in args else None

    # NOTE(ivasilev) argparse choices and defaults in enough for validation
    report_schema = args.report_schema

    if os.getuid():
        raise CommandError('This command has to be run under the root user.')

    if args.resume:
        context, configuration = util.fetch_last_upgrade_context(resume_context)
        if not context:
            raise CommandError('No previous upgrade run to continue, remove `--resume` from leapp invocation to'
                               ' start a new upgrade flow')
        os.environ['LEAPP_DEBUG'] = '1' if util.check_env_and_conf('LEAPP_DEBUG', 'debug', configuration) else '0'

        if os.environ['LEAPP_DEBUG'] == '1' or util.check_env_and_conf('LEAPP_VERBOSE', 'verbose', configuration):
            os.environ['LEAPP_VERBOSE'] = '1'
        else:
            os.environ['LEAPP_VERBOSE'] = '0'
        util.restore_leapp_env_vars(context)
        skip_phases_until = util.get_last_phase(context)
    else:
        util.disable_database_sync()
        configuration = util.prepare_configuration(args)
        e = Execution(context=context, kind='upgrade', configuration=configuration)
        e.store()
        util.archive_logfiles()

    logger = configure_logger('leapp-upgrade.log')
    os.environ['LEAPP_EXECUTION_ID'] = context

    if args.resume:
        logger.info("Resuming execution after phase: %s", skip_phases_until)
    try:
        repositories = util.load_repositories()
    except LeappError as exc:
        raise CommandError(exc.message)
    workflow = repositories.lookup_workflow('IPUWorkflow')(auto_reboot=args.reboot)
    util.process_whitelist_experimental(repositories, workflow, configuration, logger)
    util.warn_if_unsupported(configuration)
    with beautify_actor_exception():
        logger.info("Using answerfile at %s", answerfile_path)
        workflow.load_answers(answerfile_path, userchoices_path)

        # Set the locale, so that the actors parsing command outputs that might be localized will not fail
        os.environ['LANGUAGE'] = 'en_US.UTF-8'
        os.environ['LC_ALL'] = 'en_US.UTF-8'
        os.environ['LANG'] = 'en_US.UTF-8'
        workflow.run(context=context, skip_phases_until=skip_phases_until, skip_dialogs=True,
                     only_with_tags=only_with_tags)

    logger.info("Answerfile will be created at %s", answerfile_path)
    workflow.save_answers(answerfile_path, userchoices_path)
    report_errors(workflow.errors)
    util.generate_report_files(context, report_schema)
    report_files = util.get_cfg_files('report', cfg)
    log_files = util.get_cfg_files('logs', cfg)
    report_info(context, report_files, log_files, answerfile_path, fail=workflow.failure, errors=workflow.errors)

    if workflow.failure:
        sys.exit(1)


def register(base_command):
    """
        Registers `leapp upgrade`
    """
    base_command.add_sub(upgrade)



import subprocess
from io import BytesIO

from leapp.compat import IS_PYTHON3, unicode_type
from leapp.libraries.actor import scansaphana

SAPHANA_TEST_INSTANCES = ('00', '10', '23', '34', '66')
SAPHANA2_MANIFEST = ''.join(('''compversion-id: 73554900100200005327
comptype: HDB
keyname: HDB
keycaption: SAP HANA DATABASE
supported-phases: prepare,offline,configure,online
requires-restart: system
keyvendor: sap.com
release: 2.00
rev-number: 053
rev-patchlevel: 00
rev-changelist: 1605092543
max_sps12_rev-number: 122
max_sps12_patchlevel: 33
max_rel2.0_sps0_rev-number: 02
max_rel2.0_sps0_patchlevel: 02
max_rel2.0_sps1_rev-number: 12
max_rel2.0_sps1_patchlevel: 05
max_rel2.0_sps2_rev-number: 24
max_rel2.0_sps2_patchlevel: 10
max_rel2.0_sps3_rev-number: 37
max_rel2.0_sps3_patchlevel: 07
max_rel2.0_sps4_rev-number: 48
max_rel2.0_sps4_patchlevel: 02
upgrade-restriction: sourceVersion="[1.00.000.00,1.00.69.07)"; message="You must first upgrade your system to the ''',
                             '''latest HANA 1.0 SPS6 Revision (69.07), then to HANA 1.0 SPS12 if you want to ''',
                             '''upgrade to HANA 2.0"
upgrade-restriction: sourceVersion="[1.00.69.07,1.00.120.00)"; message="You must first make an intermediate update ''',
                             '''to HANA SPS12 if you want to upgrade to HANA 2.0. Please see SAP Note 2372809."
restrict-supported-components: name="REMOTE_DATA_SYNC"; vendor="sap.com"; version="[2.0.010.00,3.00.000.00)"
sp-number: 053
sp-patchlevel: 00
makeid: 7694334
date: 2020-11-11 12:12:22
platform: linuxx86_64
hdb-state: RAMP
fullversion: 2.00.053.00 Build 1605092543-1530
auxversion: 0000.00.0
cloud_edition: 0000.00.00
changeinfo: CONT 49fd3e766fbee9a5c22dd609397d2fa640b9df09 (fa/hana2sp05)
compiletype: rel
compilebranch: fa/hana2sp05
git-hash: 49fd3e766fbee9a5c22dd609397d2fa640b9df09
git-headcount: 500015
git-mergetime: 2020-11-11 12:02:23
git-mergeepoch: 1605092543
sapexe-version: 753
sapexe-branch: 753_REL
sapexe-changelist: 2007209
compiler-version-full: gcc (SAP release 20200227, based on SUSE gcc9-9.2.1+r275327-1.3.7) 9.2.1 20190903 ''',
                             '''[gcc-9-branch revision 275330]
compiler-version: GCC 9
lcmserver-artifact-version: 2.5.46'''))


class CallMock(object):
    def __init__(self, ret):
        self.args = None
        self.ret = ret

    def __call__(self, *args, **kwargs):
        self.args = args
        return self.ret


class SubprocessCall(object):
    def __init__(self, admusername):
        self.admusername = admusername

    def __call__(self, *args, **kwargs):
        assert args[0][0:3] == ['sudo', '-u', self.admusername]
        cmd = args[0][3:]
        kwargs.pop('checked', None)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        p.wait()
        return {'exit_code': p.returncode, 'stdout': p.stdout.read()}


def test_scansaphana_get_instance_status(monkeypatch):
    call = CallMock(ret={})
    monkeypatch.setattr(scansaphana, 'run', call)

    call.ret = {'exit_code': 3, 'stdout': ''}
    assert scansaphana.get_instance_status('00', 'fake/control/path', 'tstadm')
    assert call.args[0] == ['sudo', '-u', 'tstadm', 'fake/control/path', '-nr', '00', '-function', 'GetProcessList']

    call.ret = {'exit_code': 4, 'stdout': ''}
    assert not scansaphana.get_instance_status('00', 'fake/control/path', 'tstadm')
    assert call.args[0] == ['sudo', '-u', 'tstadm', 'fake/control/path', '-nr', '00', '-function', 'GetProcessList']

    call.ret = {'exit_code': 0, 'stdout': '\n'}
    assert not scansaphana.get_instance_status('00', 'fake/control/path', 'tstadm')
    assert call.args[0] == ['sudo', '-u', 'tstadm', 'fake/control/path', '-nr', '00', '-function', 'GetProcessList']

    call.ret = {'exit_code': 0, 'stdout': ' \n' * 6}
    assert not scansaphana.get_instance_status('00', 'fake/control/path', 'tstadm')
    assert call.args[0] == ['sudo', '-u', 'tstadm', 'fake/control/path', '-nr', '00', '-function', 'GetProcessList']

    call.ret = {'exit_code': 0, 'stdout': '\n' * 7}
    assert scansaphana.get_instance_status('00', 'fake/control/path', 'tstadm')
    assert call.args[0] == ['sudo', '-u', 'tstadm', 'fake/control/path', '-nr', '00', '-function', 'GetProcessList']


def test_scansaphana_parse_manifest(monkeypatch):
    class _mock_open(object):
        def __init__(self, path, mode):
            self._fp = BytesIO(SAPHANA2_MANIFEST.encode('utf-8'))

        def __enter__(self):
            return self._fp

        def __exit__(self, *args, **kwargs):
            return None

    monkeypatch.setattr(scansaphana, 'open', _mock_open, False)
    data = scansaphana.parse_manifest('yadda')

    def _get_manifest_entry(key):
        for entry in data:
            if entry.key == key:
                return entry.value
        return None
    assert _get_manifest_entry('sapexe-version') == '753'
    assert _get_manifest_entry('keycaption') == 'SAP HANA DATABASE'


def test_scansaphana_search_saphana_databases(monkeypatch, tmpdir):
    base_path = tmpdir
    monkeypatch.setattr(scansaphana, 'HANA_BASE_PATH', str(base_path))
    cur_path = base_path
    manifest_path_parts = scansaphana.HANA_MANIFEST_PATH.split('/')
    for part in manifest_path_parts[0:-1]:
        cur_path = cur_path.join(part)
        cur_path.mkdir()

    linkpath = base_path.join(manifest_path_parts[0])

    manifest = cur_path.join(manifest_path_parts[-1])
    manifest.write_text(unicode_type(SAPHANA2_MANIFEST), encoding='utf-8')

    sapcontrol = cur_path.join(scansaphana.HANA_SAPCONTROL_PATH.split('/')[-1])
    sapcontrol.write_text(unicode_type('''#!/bin/bash
echo ''
echo '26.01.2021 17:55:17'
echo 'GetProcessList'
echo 'OK'
echo 'name, description, dispstatus, textstatus, starttime, elapsedtime, pid'
echo 'hdbdaemon, HDB Daemon, GREEN, Running, 2020 05 08 21:56:28, 6307:58:49, 16043'
echo 'hdbcompileserver, HDB Compileserver, GREEN, Running, 2020 05 08 21:57:00, 6307:58:17, 16365'
echo 'hdbdiserver, HDB Deployment Infrastructure Server, GREEN, Running, 2020 05 08 21:57:37, 6307:57:40, 16546'
echo 'hdbindexserver, HDB Indexserver-HXE, GREEN, Running, 2020 05 08 21:57:01, 6307:58:16, 16391'
echo 'hdbnameserver, HDB Nameserver, GREEN, Running, 2020 05 08 21:56:29, 6307:58:48, 16061'
echo 'hdbwebdispatcher, HDB Web Dispatcher, GREEN, Running, 2020 05 08 21:57:38, 6307:57:39, 16549'
exit 3
    '''), encoding='utf-8')
    sapcontrol.chmod(0o755)

    result = scansaphana.search_sap_hana_instances()
    assert not result.instances
    assert not result.running
    assert not result.installed

    monkeypatch.setattr(scansaphana, 'run', SubprocessCall('lppadm'))
    admin = base_path.join('LPP')
    admin.mkdir()
    admin.join(manifest_path_parts[0]).mksymlinkto(str(linkpath))

    for instance in SAPHANA_TEST_INSTANCES:
        instance_path = admin.join(('HDB' + instance))
        instance_path.mkdir()

    result = scansaphana.search_sap_hana_instances()
    assert result.instances
    assert len(result.instances) == len(SAPHANA_TEST_INSTANCES)
    check_instance_numbers = set(SAPHANA_TEST_INSTANCES)
    for instance in result.instances:
        assert instance.instance_number in check_instance_numbers
        check_instance_numbers.remove(instance.instance_number)
    assert not check_instance_numbers
    assert result.running
    assert result.installed



from leapp.models import OpenSslConfig, OpenSslConfigBlock, OpenSslConfigPair, Report


def test_actor_execution_empty(current_actor_context):
    current_actor_context.feed(
        OpenSslConfig(
            blocks=[],
            # modified=False, # default
        )
    )
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_execution_empty_modified(current_actor_context):
    current_actor_context.feed(
        OpenSslConfig(
            blocks=[],
            modified=True,
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert 'missing the crypto policies integration' in r[0].report['title']


def test_actor_execution_default_modified(current_actor_context):
    current_actor_context.feed(
        OpenSslConfig(
            openssl_conf='default_modules',
            blocks=[
                OpenSslConfigBlock(
                    name="default_modules",
                    pairs=[
                        OpenSslConfigPair(
                            key="ssl_conf",
                            value="ssl_module"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="ssl_module",
                    pairs=[
                        OpenSslConfigPair(
                            key="system_default",
                            value="crypto_policy"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="crypto_policy",
                    pairs=[
                        OpenSslConfigPair(
                            key=".include",
                            value="/etc/crypto-policies/back-ends/opensslcnf.config"
                        )
                    ]
                ),
            ],
            modified=True,
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert 'The OpenSSL configuration will be updated to support OpenSSL 3.0' in r[0].report['title']


def test_actor_execution_minprotocol_modified(current_actor_context):
    current_actor_context.feed(
        OpenSslConfig(
            openssl_conf='default_modules',
            blocks=[
                OpenSslConfigBlock(
                    name="default_modules",
                    pairs=[
                        OpenSslConfigPair(
                            key="ssl_conf",
                            value="ssl_module"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="ssl_module",
                    pairs=[
                        OpenSslConfigPair(
                            key="system_default",
                            value="crypto_policy"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="crypto_policy",
                    pairs=[
                        OpenSslConfigPair(
                            key=".include",
                            value="/etc/crypto-policies/back-ends/opensslcnf.config"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="crypto_policy",
                    pairs=[
                        OpenSslConfigPair(
                            key="TLS.MinProtocol",
                            value="TLSv1.2"
                        ),
                    ]
                )
            ],
            modified=True,
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert 'MinProtocol and MaxProtocol' in r[0].report['title']
    assert 'The OpenSSL configuration will be updated to support OpenSSL 3.0' in r[1].report['title']


def test_actor_execution_duplicate_extensions_modified(current_actor_context):
    current_actor_context.feed(
        OpenSslConfig(
            openssl_conf='default_modules',
            blocks=[
                OpenSslConfigBlock(
                    name="default_modules",
                    pairs=[
                        OpenSslConfigPair(
                            key="ssl_conf",
                            value="ssl_module"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="ssl_module",
                    pairs=[
                        OpenSslConfigPair(
                            key="system_default",
                            value="crypto_policy"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="crypto_policy",
                    pairs=[
                        OpenSslConfigPair(
                            key=".include",
                            value="/etc/crypto-policies/back-ends/opensslcnf.config"
                        )
                    ]
                ),
                OpenSslConfigBlock(
                    name="req",
                    pairs=[
                        OpenSslConfigPair(
                            key="x509_extensions",
                            value="v3_ca",
                        ),
                    ]
                ),
                OpenSslConfigBlock(
                    name="v3_ca",
                    pairs=[
                        OpenSslConfigPair(
                            key="keyUsage",
                            value="cRLSign, keyCertSign",
                        ),
                        OpenSslConfigPair(
                            key="keyUsage",
                            value="cRLSign",
                        ),  # yay, duplicate
                        OpenSslConfigPair(
                            key="keyUsage",
                            value="keyCertSign",
                        ),  # yay, duplicate
                    ]
                ),
            ],
            modified=True,
        )
    )
    current_actor_context.run()
    r = current_actor_context.consume(Report)
    assert r
    assert 'duplicate x509 extensions' in r[0].report['title']
    assert 'keyUsage' in r[0].report['summary']
    assert 'The OpenSSL configuration will be updated to support OpenSSL 3.0' in r[1].report['title']



from leapp.libraries.actor.tcpwrapperscheck import config_affects_daemons
from leapp.models import DaemonList, TcpWrappersFacts


def test_empty_packages():
    """ Test with empty package list """
    w = TcpWrappersFacts(daemon_lists=[DaemonList(value=["ALL"])])
    p = []
    d = [("openssh", ["sshd"])]
    packages = config_affects_daemons(w, p, d)

    assert not packages


def test_empty_tcp_wrappers():
    """ Test with empty tcp_wrappers facts """
    w = TcpWrappersFacts(daemon_lists=[])
    p = ["openssh", "systemd", "pam"]
    d = [("openssh", ["sshd"])]
    packages = config_affects_daemons(w, p, d)

    assert not packages


def test_matching_package():
    """ Test with matching package, but not daemon """
    w = TcpWrappersFacts(daemon_lists=[DaemonList(value=["imap"])])
    p = ["openssh", "systemd", "pam"]
    d = [("openssh", ["sshd"])]
    packages = config_affects_daemons(w, p, d)

    assert not packages


def test_matching_daemon():
    """ Test with matching package with daemon """
    w = TcpWrappersFacts(daemon_lists=[DaemonList(value=["sshd"])])
    p = ["openssh", "systemd", "pam"]
    d = [("openssh", ["sshd"])]
    packages = config_affects_daemons(w, p, d)

    assert len(packages) == 1
    assert "openssh" in packages


def test_matching_all():
    """ Test with matching package with daemon """
    w = TcpWrappersFacts(daemon_lists=[DaemonList(value=["ALL"])])
    p = ["openssh", "systemd", "pam", "audit"]
    d = [("openssh", ["sshd"]), ("audit", ["auditd"])]
    packages = config_affects_daemons(w, p, d)

    assert len(packages) == 2
    assert "openssh" in packages
    assert "audit" in packages



import pytest

from leapp import reporting
from leapp.libraries.actor.checkyumpluginsenabled import check_required_yum_plugins_enabled
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import PkgManagerInfo
from leapp.utils.report import is_inhibitor


def _create_report_with_multiple_base_list_primitive_fields():
    report_fields = [
            reporting.Title('a title'),
            reporting.Summary('some summary'),
            reporting.RelatedResource('file', '/path/to/some/configA'),
            reporting.RelatedResource('file', '/path/to/some/configB'),
            reporting.RelatedResource('file', '/path/to/some/configC'),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.Groups([reporting.Groups.REPOSITORY])]
    reporting.create_report(report_fields)


def test__create_report_mocked(monkeypatch):
    actor_reports = create_report_mocked()
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(reporting, 'create_report', actor_reports)
    _create_report_with_multiple_base_list_primitive_fields()

    # make sure that all related resources are present
    for config in ['/path/to/some/configA', '/path/to/some/configB', '/path/to/some/configC']:
        assert config in [rr['title'] for rr in actor_reports.report_fields['detail']['related_resources']]

    # make sure that tags/flags joined as groups are present under groups
    for group in [reporting.Groups.INHIBITOR, reporting.Groups.REPOSITORY]:
        assert group in actor_reports.report_fields['groups']


def test_report_when_missing_required_plugins(monkeypatch):
    """Test whether a report entry is created when any of the required YUM plugins are missing."""
    yum_config = PkgManagerInfo(enabled_plugins=['product-id', 'some-user-plugin'])

    actor_reports = create_report_mocked()

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(reporting, 'create_report', actor_reports)

    check_required_yum_plugins_enabled(yum_config)

    assert actor_reports.called, "Report wasn't created when required a plugin is missing."

    fail_description = 'The missing required plugin is not mentioned in the report.'
    assert 'subscription-manager' in actor_reports.report_fields['summary'], fail_description

    fail_description = 'The upgrade should be inhibited when plugins are not enabled.'
    assert is_inhibitor(actor_reports.report_fields), fail_description


def test_nothing_is_reported_when_rhsm_disabled(monkeypatch):
    actor_mocked = CurrentActorMocked(envars={'LEAPP_NO_RHSM': '1'})

    monkeypatch.setattr(api, 'current_actor', actor_mocked)
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    yum_config = PkgManagerInfo(enabled_plugins=[])
    check_required_yum_plugins_enabled(yum_config)

    assert not reporting.create_report.called, 'Report was created even if LEAPP_NO_RHSM was set'



import contextlib
import os

from six.moves.urllib.error import URLError
from six.moves.urllib.request import urlopen

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import CalledProcessError


@contextlib.contextmanager
def guarded_execution(*guards):
    try:
        yield
    except CalledProcessError as e:
        # collect output from guards for possible spurious failure
        guard_errors = []
        for guard in guards:
            err = guard()
            if err:
                guard_errors.append(err)

        details = None
        if guard_errors:
            details = {
                'hint': 'Possible spurious failure: {cause}'.format(cause=' '.join(guard_errors))
            }
        raise StopActorExecutionError(
            message=str(e),
            details=details
        )


def connection_guard(url='https://example.com'):
    def closure():
        try:
            urlopen(url)
            return None
        except URLError as e:
            cause = '''Failed to open url '{url}' with error: {error}'''.format(url=url, error=e)
            return ('There was probably a problem with internet connection ({cause}).'
                    ' Check your connection and try again.'.format(cause=cause))
    return closure


def space_guard(path='/', min_free_mb=100):
    def closure():
        info = os.statvfs(path)
        free_mb = (info.f_bavail * info.f_frsize) >> 20
        if free_mb >= min_free_mb:
            return None
        return ('''Not enough free disk space in '{path}', needed: {min} M, available: {avail} M.'''
                ' Free more disk space and try again.'.format(path=path, min=min_free_mb, avail=free_mb))
    return closure



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import check_fips
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import FIPSInfo


@pytest.mark.parametrize(
    ('fips_info', 'sys_fips_enabled_contents', 'should_prevent_ipu'),
    (
        (FIPSInfo(is_enabled=False), '0', False),
        (FIPSInfo(is_enabled=True), '0', True),
        (FIPSInfo(is_enabled=True), '1', False),
    )
)
def test_ipu_prevention_if_fips_not_perserved(monkeypatch,
                                              fips_info,
                                              sys_fips_enabled_contents,
                                              should_prevent_ipu):

    mocked_actor = CurrentActorMocked(msgs=[fips_info])
    monkeypatch.setattr(check_fips, 'read_sys_fips_state', lambda: sys_fips_enabled_contents)
    monkeypatch.setattr(api, 'current_actor', mocked_actor)

    if should_prevent_ipu:
        with pytest.raises(StopActorExecutionError):
            check_fips.check_fips_state_perserved()
    else:
        check_fips.check_fips_state_perserved()  # unhandled exception with crash the test



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class SystemdBrokenSymlinksSource(Model):
    """
    Information about broken systemd symlinks on the source system
    """

    topic = SystemInfoTopic
    broken_symlinks = fields.List(fields.String(), default=[])
    """
    List of broken systemd symlinks on the source system

    The values are absolute paths of the broken symlinks.
    """


class SystemdBrokenSymlinksTarget(SystemdBrokenSymlinksSource):
    """
    Analogy to :class:`SystemdBrokenSymlinksSource`, but for the target system
    """


class SystemdServicesTasks(Model):
    """
    Influence the systemd services of the target system

    E.g. it could be specified explicitly whether some services should
    be enabled or disabled after the in-place upgrade - follow descriptions
    of particular tasks for details.

    In case of conflicting tasks (e.g. the A service should be enabled and
    disabled in the same time):
       a) If conflicting tasks are detected during check phases,
          the upgrade is inhibited with the proper report.
       b) If conflicting tasks are detected during the final evaluation,
          error logs are created and such services will be disabled.
    """
    topic = SystemInfoTopic

    to_enable = fields.List(fields.String(), default=[])
    """
    List of systemd services to enable on the target system

    Masked services will not be enabled. Attempting to enable a masked service
    will be evaluated by systemctl as usually. The error will be logged and the
    upgrade process will continue.
    """

    to_disable = fields.List(fields.String(), default=[])
    """
    List of systemd services to disable on the target system
    """

    # NOTE: possible extension in case of requirement (currently not implemented):
    # to_unmask = fields.List(fields.String(), default=[])


class SystemdServiceFile(Model):
    """
    Information about single systemd service unit file

    This model is not expected to be produced nor consumed by actors directly.
    See the :class:`SystemdServicesInfoSource` and :class:`SystemdServicesPresetInfoTarget`
    for more info.
    """
    topic = SystemInfoTopic

    name = fields.String()
    """
    Name of the service unit file
    """

    state = fields.StringEnum([
        'alias',
        'bad',
        'disabled',
        'enabled',
        'enabled-runtime',
        'generated',
        'indirect',
        'linked',
        'linked-runtime',
        'masked',
        'masked-runtime',
        'static',
        'transient',
    ])
    """
    The state of the service unit file
    """


class SystemdServicesInfoSource(Model):
    """
    Information about systemd services on the source system
    """
    topic = SystemInfoTopic

    service_files = fields.List(fields.Model(SystemdServiceFile), default=[])
    """
    List of all installed systemd service unit files

    Instances of service template unit files don't have a unit file
    and therefore aren't included, but their template files are.
    Generated service unit files are also included.
    """


class SystemdServicesInfoTarget(SystemdServicesInfoSource):
    """
    Analogy to :class:`SystemdServicesInfoSource`, but for the target system

    This information is taken after the RPM Upgrade and might become
    invalid if there are actors calling systemctl enable/disable directly later
    in the upgrade process. Therefore it is recommended to use
    :class:`SystemdServicesTasks` to alter the state of units in the
    FinalizationPhase.
    """


class SystemdServicePreset(Model):
    """
    Information about a preset for systemd service
    """

    topic = SystemInfoTopic
    service = fields.String()
    """
    Name of the service, with the .service suffix
    """

    state = fields.StringEnum(['disable', 'enable'])
    """
    The state set by a preset file
    """


class SystemdServicesPresetInfoSource(Model):
    """
    Information about presets for systemd services
    """
    topic = SystemInfoTopic

    presets = fields.List(fields.Model(SystemdServicePreset), default=[])
    """
    List of all service presets
    """


class SystemdServicesPresetInfoTarget(SystemdServicesPresetInfoSource):
    """
    Analogy to :class:`SystemdServicesPresetInfoSource` but for the target system
    """



from leapp.models import CephInfo, LsblkEntry, StorageInfo
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


def test_actor_with_luks(current_actor_context):
    with_luks = [LsblkEntry(name='luks-132', kname='kname1', maj_min='253:0', rm='0',
                            size='10G', bsize=10*(1 << 39), ro='0', tp='crypt', mountpoint='')]

    current_actor_context.feed(StorageInfo(lsblk=with_luks))
    current_actor_context.run()
    assert current_actor_context.consume(Report)
    report_fields = current_actor_context.consume(Report)[0].report
    assert is_inhibitor(report_fields)


def test_actor_with_luks_ceph_only(current_actor_context):
    with_luks = [LsblkEntry(name='luks-132', kname='kname1', maj_min='253:0', rm='0',
                            size='10G', bsize=10*(1 << 39), ro='0', tp='crypt', mountpoint='')]
    ceph_volume = ['luks-132']
    current_actor_context.feed(StorageInfo(lsblk=with_luks))
    current_actor_context.feed(CephInfo(encrypted_volumes=ceph_volume))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_without_luks(current_actor_context):
    without_luks = [LsblkEntry(name='sda1', kname='sda1', maj_min='8:0', rm='0',
                               size='10G', bsize=10*(1 << 39), ro='0', tp='part', mountpoint='/boot')]

    current_actor_context.feed(StorageInfo(lsblk=without_luks))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



import os
import re

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import BootContent, KernelCmdlineArg, TargetKernelCmdlineArgTasks


def add_boot_entry(configs=None):
    debug = 'debug' if os.getenv('LEAPP_DEBUG', '0') == '1' else ''
    enable_network = os.getenv('LEAPP_DEVEL_INITRAM_NETWORK') in ('network-manager', 'scripts')
    ip_arg = ' ip=dhcp rd.neednet=1' if enable_network else ''
    kernel_dst_path, initram_dst_path = get_boot_file_paths()
    _remove_old_upgrade_boot_entry(kernel_dst_path, configs=configs)
    try:
        cmd = [
            '/usr/sbin/grubby',
            '--add-kernel', '{0}'.format(kernel_dst_path),
            '--initrd', '{0}'.format(initram_dst_path),
            '--title', 'RHEL-Upgrade-Initramfs',
            '--copy-default',
            '--make-default',
            '--args', '{DEBUG}{NET} enforcing=0 rd.plymouth=0 plymouth.enable=0'.format(DEBUG=debug, NET=ip_arg)
        ]
        if configs:
            for config in configs:
                run(cmd + ['-c', config])
        else:
            run(cmd)

        if architecture.matches_architecture(architecture.ARCH_S390X):
            # on s390x we need to call zipl explicitly because of issue in grubby,
            # otherwise the new boot entry will not be set as default
            # See https://bugzilla.redhat.com/show_bug.cgi?id=1764306
            run(['/usr/sbin/zipl'])

        if debug:
            # The kernelopts for target kernel are generated based on the cmdline used in the upgrade initramfs,
            # therefore, if we enabled debug above, and the original system did not have the debug kernelopt, we
            # need to explicitly remove it from the target os boot entry.
            # NOTE(mhecko): This will also unconditionally remove debug kernelopt if the source system used it.
            api.produce(TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='debug')]))

        # NOTE(mmatuska): This will remove the option even if the source system had it set.
        # However enforcing=0 shouldn't be set persistently anyway.
        api.produce(TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='enforcing', value='0')]))

    except CalledProcessError as e:
        raise StopActorExecutionError(
           'Cannot configure bootloader.',
           details={'details': '{}: {}'.format(str(e), e.stderr)}
        )


def _remove_old_upgrade_boot_entry(kernel_dst_path, configs=None):
    """
    Remove entry referring to the upgrade kernel.

    We have to ensure there are no duplicit boot entries. Main reason is crash
    of zipl when duplicit entries exist.
    """
    cmd = [
        '/usr/sbin/grubby',
        '--remove-kernel', '{0}'.format(kernel_dst_path)
    ]
    try:
        if configs:
            for config in configs:
                run(cmd + ['-c', config])
        else:
            run(cmd)
    except CalledProcessError:
        # TODO(pstodulk): instead of this, check whether the entry exists or not
        # so no warning of problem is reported (info log could be present if the
        # entry is missing.
        api.current_logger().warning(
            'Could not remove {} entry. May be ignored if the entry did not exist.'.format(kernel_dst_path)
        )


def get_boot_file_paths():
    boot_content_msgs = api.consume(BootContent)
    boot_content = next(boot_content_msgs, None)
    if list(boot_content_msgs):
        api.current_logger().warning('Unexpectedly received more than one BootContent message.')
    if not boot_content:
        raise StopActorExecutionError('Could not create a GRUB boot entry for the upgrade initramfs',
                                      details={'details': 'Did not receive a message about the leapp-provided'
                                                          'kernel and initramfs'})
    # Returning information about kernel hmac file path is needless as it is not used when adding boot entry
    return boot_content.kernel_path, boot_content.initram_path


def write_to_file(filename, content):
    with open(filename, 'w') as f:
        f.write(content)


def fix_grub_config_error(conf_file, error_type):
    with open(conf_file, 'r') as f:
        config = f.read()

    if error_type == 'GRUB_CMDLINE_LINUX syntax':
        # move misplaced '"' to the end
        pattern = r'GRUB_CMDLINE_LINUX=.+?(?=GRUB|\Z)'
        original_value = re.search(pattern, config, re.DOTALL).group()
        parsed_value = original_value.split('"')
        new_value = '{KEY}"{VALUE}"{END}'.format(KEY=parsed_value[0], VALUE=''.join(parsed_value[1:]).rstrip(),
                                                 END=original_value[-1])

        config = config.replace(original_value, new_value)
        write_to_file(conf_file, config)

    elif error_type == 'missing newline':
        write_to_file(conf_file, config + '\n')



from leapp.actors import Actor
from leapp.libraries.actor import checkrhui as checkrhui_lib
from leapp.models import (
    CopyFile,
    DNFPluginTask,
    InstalledRPM,
    KernelCmdlineArg,
    RequiredTargetUserspacePackages,
    RHUIInfo,
    RpmTransactionTasks,
    TargetUserSpacePreupgradeTasks
)
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class CheckRHUI(Actor):
    """
    Check if system is using RHUI infrastructure (on public cloud) and send messages to
    provide additional data needed for upgrade.
    """

    name = 'checkrhui'
    consumes = (InstalledRPM,)
    produces = (
        KernelCmdlineArg,
        RHUIInfo,
        RequiredTargetUserspacePackages,
        Report, DNFPluginTask,
        RpmTransactionTasks,
        TargetUserSpacePreupgradeTasks,
        CopyFile,
    )
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        checkrhui_lib.process()



import os

import pytest

from leapp.models import SelinuxPermissiveDecision


@pytest.mark.skipif(
    os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
    reason='Test disabled by default because it would modify the system')
def check_permissive_in_conf():
    """ Check if we have set permissive in SElinux conf file """
    with open('/etc/selinux/config') as fo:
        result = [l for l in (line.strip() for line in fo) if l]
        for res in result:
            if res == "SELINUX=permissive":
                return True
    return False


@pytest.mark.skipif(
    os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
    reason='Test disabled by default because it would modify the system')
def test_set_selinux_permissive(current_actor_context):
    current_actor_context.feed(SelinuxPermissiveDecision(set_permissive=True))
    current_actor_context.run()
    assert check_permissive_in_conf()



from collections import namedtuple

import pytest

from leapp.libraries import stdlib
from leapp.libraries.actor import kernelcmdlineconfig
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelInfo, KernelCmdlineArg, TargetKernelCmdlineArgTasks

TARGET_KERNEL_NEVRA = 'kernel-core-1.2.3-4.x86_64.el8.x64_64'


class MockedRun(object):
    def __init__(self):
        self.commands = []

    def __call__(self, cmd, *args, **kwargs):
        self.commands.append(cmd)
        return {}


@pytest.mark.parametrize(
    ('msgs', 'expected_grubby_kernelopt_args'),
    [
        (
            [KernelCmdlineArg(key='key1', value='value1'), KernelCmdlineArg(key='key2', value='value2')],
            ['--args', 'key1=value1 key2=value2']
        ),
        (
            [TargetKernelCmdlineArgTasks(to_add=[KernelCmdlineArg(key='key1', value='value1'),
                                                 KernelCmdlineArg(key='key2')])],
            ['--args', 'key1=value1 key2']
        ),
        (
            [TargetKernelCmdlineArgTasks(to_add=[KernelCmdlineArg(key='key1', value='value1')]),
             KernelCmdlineArg(key='key2', value='value2')],
            ['--args', 'key1=value1 key2=value2']
        ),
        (
            [TargetKernelCmdlineArgTasks(to_add=[KernelCmdlineArg(key='key1', value='value1')],
                                         to_remove=[KernelCmdlineArg(key='key3')]),
             KernelCmdlineArg(key='key2', value='value2')],
            ['--args', 'key1=value1 key2=value2', '--remove-args', 'key3']
        ),
        (
            [TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='key3'), KernelCmdlineArg(key='key4')])],
            ['--remove-args', 'key3 key4']
        ),
    ]
)
def test_kernelcmdline_config_valid_msgs(monkeypatch, msgs, expected_grubby_kernelopt_args):
    kernel_img_path = '/boot/vmlinuz-X'
    kernel_info = InstalledTargetKernelInfo(pkg_nevra=TARGET_KERNEL_NEVRA,
                                            uname_r='',
                                            kernel_img_path=kernel_img_path,
                                            initramfs_path='/boot/initramfs-X')
    msgs += [kernel_info]

    grubby_base_cmd = ['grubby', '--update-kernel={}'.format(kernel_img_path)]
    expected_grubby_cmd = grubby_base_cmd + expected_grubby_kernelopt_args

    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_X86_64, msgs=msgs))
    kernelcmdlineconfig.modify_kernel_args_in_boot_cfg()
    assert mocked_run.commands and len(mocked_run.commands) == 1
    assert expected_grubby_cmd == mocked_run.commands.pop()

    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X, msgs=msgs))
    kernelcmdlineconfig.modify_kernel_args_in_boot_cfg()
    assert mocked_run.commands and len(mocked_run.commands) == 2
    assert expected_grubby_cmd == mocked_run.commands.pop(0)
    assert ['/usr/sbin/zipl'] == mocked_run.commands.pop(0)


def test_kernelcmdline_explicit_configs(monkeypatch):
    kernel_img_path = '/boot/vmlinuz-X'

    kernel_info = InstalledTargetKernelInfo(pkg_nevra=TARGET_KERNEL_NEVRA,
                                            uname_r='',
                                            kernel_img_path=kernel_img_path,
                                            initramfs_path='/boot/initramfs-X')
    msgs = [kernel_info, TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='key1', value='value1')])]

    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_X86_64, msgs=msgs))

    configs = ['/boot/grub2/grub.cfg', '/boot/efi/EFI/redhat/grub.cfg']
    kernelcmdlineconfig.modify_kernel_args_in_boot_cfg(configs_to_modify_explicitly=configs)

    grubby_cmd_without_config = ['grubby', '--update-kernel={}'.format(kernel_img_path),
                                 '--remove-args', 'key1=value1']
    expected_cmds = [
        grubby_cmd_without_config + ['-c', '/boot/grub2/grub.cfg'],
        grubby_cmd_without_config + ['-c', '/boot/efi/EFI/redhat/grub.cfg']
    ]

    assert mocked_run.commands == expected_cmds


def test_kernelcmdline_config_no_args(monkeypatch):
    kernel_info = InstalledTargetKernelInfo(pkg_nevra=TARGET_KERNEL_NEVRA,
                                            uname_r='',
                                            kernel_img_path='/boot/vmlinuz-X',
                                            initramfs_path='/boot/initramfs-X')

    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X, msgs=[kernel_info]))
    kernelcmdlineconfig.modify_kernel_args_in_boot_cfg()
    assert not mocked_run.commands


def test_kernelcmdline_config_no_version(monkeypatch):
    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(architecture.ARCH_S390X))
    kernelcmdlineconfig.modify_kernel_args_in_boot_cfg()
    assert not mocked_run.commands



from leapp.actors import Actor
from leapp.dialogs import Dialog
from leapp.dialogs.components import BooleanComponent
from leapp.libraries.actor.checkvdo import check_vdo
from leapp.models import Report, VdoConversionInfo
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckVdo(Actor):
    """
    Check if VDO devices need to be migrated to lvm management.

    `Background`
    ============
    In RHEL 9.0 the independent VDO management software, `vdo manager`, is
    superseded by LVM management.  Existing VDOs must be converted to LVM-based
    management *before* upgrading to RHEL 9.0.

    The `CheckVdo` actor provides a pre-upgrade check for VDO devices that have
    not been converted and, if any are found, produces an inhibitory report
    during upgrade check phase.   If none are found `CheckVdo` does not produce
    a report concerning such VDO devices.

    As there currently exists the theoretical possibility that a VDO device may
    not complete its conversion to LVM-based management (e.g., via a poorly
    timed system crash during the conversion) `CheckVdo` also provides a
    pre-upgrade check for VDO devices in this state.  If any are found
    `CheckVdo` produces an inhibitory report during upgrade check phase.  If
    none are found `CheckVdo` does not produce a report concerning such VDO
    devices.

    If the VdoConversionInfo model indicates unexpected errors occurred during
    scanning CheckVdo will produce appropriate inhibitory reports.

    If the VdoConversionInfo model indicates conditions exist where VDO devices
    could exist but the necessary software to check was not installed on the
    system CheckVdo will present a dialog to the user. This dialog will ask the
    user to either install the required software if the user knows or is unsure
    that VDO devices exist or to approve the continuation of the upgrade if the
    user is certain that either there are no VDO devices present or that all
    VDO devices have been successfully converted.

    To maximize safety CheckVdo operates against all block devices which
    match the criteria for potential VDO devices.  Given the dynamic nature
    of device presence within a system some devices which may have been present
    during leapp discovery may not be present when CheckVdo runs.  As CheckVdo
    defaults to producing inhibitory reports if a device cannot be checked
    (for any reason) this dynamism may be problematic.  To prevent CheckVdo
    producing an inhibitory report for devices which are dynamically no longer
    present within the system the user may answer the previously mentioned
    dialog in the affirmative when the user knows that all VDO devices have
    been converted.  This will circumvent checks of block devices.
    """

    name = 'check_vdo'
    consumes = (VdoConversionInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)
    dialogs = (
        Dialog(
            scope='check_vdo',
            reason='Confirmation',
            components=(
                BooleanComponent(
                    key='confirm',
                    label='Are all VDO devices, if any, successfully converted to LVM management?',
                    description='Enter True if no VDO devices are present '
                                'on the system or all VDO devices on the system '
                                'have been successfully converted to LVM '
                                'management. '
                                'Entering True will circumvent check of failures '
                                'and undetermined devices. '
                                'Recognized VDO devices that have not been '
                                'converted to LVM management can still block '
                                'the upgrade despite the answer.'
                                'All VDO devices must be converted to LVM '
                                'management before upgrading.',
                    reason='To maximize safety all block devices on a system '
                           'that meet the criteria as possible VDO devices '
                           'are checked to verify that, if VDOs, they have '
                           'been converted to LVM management. '
                           'If the devices are not converted and the upgrade '
                           'proceeds the data on unconverted VDO devices will '
                           'be inaccessible. '
                           'In order to perform checking the \'vdo\' package '
                           'must be installed. '
                           'If the \'vdo\' package is not installed and there '
                           'are any doubts the \'vdo\' package should be '
                           'installed and the upgrade process re-run to check '
                           'for unconverted VDO devices. '
                           'If the check of any device fails for any reason '
                           'an upgrade inhibiting report is generated. '
                           'This may be problematic if devices are '
                           'dynamically removed from the system subsequent to '
                           'having been identified during device discovery. '
                           'If it is certain that all VDO devices have been '
                           'successfully converted to LVM management this '
                           'dialog may be answered in the affirmative which '
                           'will circumvent block device checking.'
                ),
            )
        ),
    )
    _asked_answer = False
    _vdo_answer = None

    def get_vdo_answer(self):
        if not self._asked_answer:
            self._asked_answer = True
            # calling this multiple times could lead to possible issues
            # or at least in redundant reports
            self._vdo_answer = self.get_answers(self.dialogs[0]).get('confirm')
        return self._vdo_answer

    def process(self):
        for conversion_info in self.consume(VdoConversionInfo):
            check_vdo(conversion_info)



from leapp.actors import Actor
from leapp.models import RpmTransactionTasks, SCTPConfig
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class SCTPChecks(Actor):
    """
    Parses collected SCTP information and take necessary actions.

    The only action performed by this actor is to request the installation of the
    kernel-modules-extra rpm package, based on if SCTP is being used or not which
    is collected on SCTPConfig message. If yes, it then produces a RpmTransactionTasks
    requesting to install the package.
    """
    name = 'sctp_checks'
    consumes = (SCTPConfig,)
    produces = (RpmTransactionTasks, )
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for sctpconfig in self.consume(SCTPConfig):
            if sctpconfig.wanted:
                self.produce(RpmTransactionTasks(to_install=['kernel-modules-extra']))
                break



import re

from leapp import reporting
from leapp.actors import Actor
from leapp.models import KernelCmdlineArg, PersistentNetNamesFacts
from leapp.reporting import create_report, Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PersistentNetNamesDisable(Actor):
    """
    Disable systemd-udevd persistent network naming on machine with single eth0 NIC
    """

    name = 'persistentnetnamesdisable'
    consumes = (PersistentNetNamesFacts,)
    produces = (KernelCmdlineArg, Report)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def ethX_count(self, interfaces):
        ethX = re.compile('eth[0-9]+')
        count = 0

        for i in interfaces:
            if ethX.match(i.name):
                count = count + 1
        return count

    def single_eth0(self, interfaces):
        return len(interfaces) == 1 and interfaces[0].name == 'eth0'

    def disable_persistent_naming(self):
        self.log.info("Single eth0 network interface detected. Appending 'net.ifnames=0' to RHEL-8 kernel commandline")
        self.produce(KernelCmdlineArg(**{'key': 'net.ifnames', 'value': '0'}))

    def process(self):
        interfaces = next(self.consume(PersistentNetNamesFacts)).interfaces

        if self.single_eth0(interfaces):
            self.disable_persistent_naming()
        elif len(interfaces) > 1 and self.ethX_count(interfaces) > 0:
            create_report([
                reporting.Title('Unsupported network configuration'),
                reporting.Summary(
                    'Detected multiple physical network interfaces where one or more use kernel naming (e.g. eth0). '
                    'Upgrade process can not continue because stability of names can not be guaranteed. '
                    'Please read the article at https://access.redhat.com/solutions/4067471 for more information.'
                ),
                reporting.ExternalLink(
                    title='How to perform an in-place upgrade to RHEL 8 when using kernel NIC names on RHEL 7',
                    url='https://access.redhat.com/solutions/4067471'
                ),
                reporting.Remediation(
                    hint='Rename all ethX network interfaces following the attached KB solution article.'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups([reporting.Groups.NETWORK]),
                reporting.Groups([reporting.Groups.INHIBITOR])
            ])



from leapp import reporting
from leapp.actors import Actor
from leapp.models import ActiveKernelModulesFacts
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckBtrfs(Actor):
    """
    Check if Btrfs filesystem is in use. If yes, inhibit the upgrade process.

    Btrfs filesystem was introduced as Technology Preview with initial releases of RHEL 6 and 7. It
    was deprecated on versions 6.6 and 7.4 and will not be present in next major version.
    """

    name = 'check_btrfs'
    consumes = (ActiveKernelModulesFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):

        hint = 'In order to unload the module from the running system, check the accompanied command.'
        command = ['modprobe', '-r', 'btrfs']

        for fact in self.consume(ActiveKernelModulesFacts):
            for active_module in fact.kernel_modules:
                if active_module.filename == 'btrfs':
                    create_report([
                        reporting.Title('Btrfs has been removed from RHEL8'),
                        reporting.Summary(
                            'The Btrfs file system was introduced as Technology Preview with the '
                            'initial release of Red Hat Enterprise Linux 6 and Red Hat Enterprise Linux 7. As of '
                            'versions 6.6 and 7.4 this technology has been deprecated and removed in RHEL8.'
                        ),
                        reporting.ExternalLink(
                            title='Considerations in adopting RHEL 8 - btrfs has been removed.',
                            url='https://red.ht/file-systems-and-storage-removed-btrfs-rhel-8'
                        ),
                        reporting.ExternalLink(
                            title='How do I prevent a kernel module from loading automatically?',
                            url='https://access.redhat.com/solutions/41278'
                        ),
                        reporting.Severity(reporting.Severity.HIGH),
                        reporting.Groups([reporting.Groups.INHIBITOR]),
                        reporting.Groups([reporting.Groups.FILESYSTEM]),
                        reporting.Remediation(hint=hint, commands=[command]),
                        reporting.RelatedResource('kernel-driver', 'btrfs')
                    ])
                    break



from leapp import reporting
from leapp.actors import Actor
from leapp.models import FirewalldGlobalConfig, FirewallsFacts
from leapp.reporting import create_report, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class FirewalldCheckAllowZoneDrifting(Actor):
    """
    This actor will check if AllowZoneDrifiting=yes in firewalld.conf. This
    option has been removed in RHEL-9 and behavior is as if
    AllowZoneDrifiting=no.
    """

    name = 'firewalld_check_allow_zone_drifting'
    consumes = (FirewallsFacts, FirewalldGlobalConfig)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        # If firewalld is not enabled then don't bother the user about its
        # configuration. This Report keys off a _default_ value and as such
        # will trigger for all users that have not done one of the following:
        #   - disabled firewalld
        #   - manually set AllowZoneDrifting=no (as firewalld logs suggests)
        #
        for facts in self.consume(FirewallsFacts):
            if not facts.firewalld.enabled:
                return

        for facts in self.consume(FirewalldGlobalConfig):
            if not facts.allowzonedrifting:
                return

        create_report([
            reporting.Title('Firewalld Configuration AllowZoneDrifting Is Unsupported'),
            reporting.Summary('Firewalld has enabled configuration option '
                              '"{conf_key}" which has been removed in RHEL-9. '
                              'New behavior is as if "{conf_key}" was set to "no".'.format(
                                  conf_key='AllowZoneDrifiting')),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SANITY, reporting.Groups.FIREWALL]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.ExternalLink(
                url='https://access.redhat.com/articles/4855631',
                title='Changes in firewalld related to Zone Drifting'),
            reporting.Remediation(
                hint='Set AllowZoneDrifting=no in /etc/firewalld/firewalld.conf',
                commands=[['sed', '-i', 's/^AllowZoneDrifting=.*/AllowZoneDrifting=no/',
                           '/etc/firewalld/firewalld.conf']]),
        ])



import os

from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.actor import migratesendmail
from leapp.libraries.stdlib import api
from leapp.models import SendmailMigrationDecision
from leapp.reporting import create_report, Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateSendmail(Actor):
    """
    Migrate sendmail configuration files.
    """

    name = 'migrate_sendmail'
    consumes = (SendmailMigrationDecision,)
    produces = (Report,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        decision = next(self.consume(SendmailMigrationDecision), None)
        if not decision or not decision.migrate_files:
            return

        not_migrated = []
        for f in decision.migrate_files:
            if not os.path.exists(f):
                api.current_logger().error('Cound not migrate file {}, because it does not exist.'.format(f))
                not_migrated.append(f)
            else:
                migratesendmail.migrate_file(f)

        list_separator_fmt = '\n    - '
        title = 'sendmail configuration files migrated'
        summary = 'Uncompressed IPv6 addresses in: {}{}'.format(list_separator_fmt,
                                                                list_separator_fmt.join(decision.migrate_files))
        severity = reporting.Severity.INFO

        if not_migrated:
            title = 'sendmail configuration files not migrated'
            summary = ('Could not migrate the configuration files, which might be caused '
                       'by removal of sendmail package during the upgrade. '
                       'Following files could not be migrated:{}{}').format(list_separator_fmt,
                                                                            list_separator_fmt.join(not_migrated))
            severity = reporting.Severity.MEDIUM

        create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(severity),
            reporting.Groups([reporting.Groups.SERVICES, reporting.Groups.EMAIL])
        ])



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class PCIAddress(Model):
    """
    TODO: tbd
    """
    topic = SystemInfoTopic

    domain = fields.String()
    bus = fields.String()
    function = fields.String()
    device = fields.String()


class Interface(Model):
    """
    TODO: tbd - Interface or NetworkInterface?
    """
    topic = SystemInfoTopic

    name = fields.String()
    devpath = fields.String()
    driver = fields.String()
    vendor = fields.String()
    pci_info = fields.Model(PCIAddress)
    mac = fields.String()


class PersistentNetNamesFacts(Model):
    """
    Information about network interfaces gather from the original system
    """
    topic = SystemInfoTopic
    interfaces = fields.List(fields.Model(Interface))


class PersistentNetNamesFactsInitramfs(PersistentNetNamesFacts):
    """
    Information about network interfaces gather from initramfs with the kernel of target system
    """
    pass


class RenamedInterface(Model):
    """
    Provide original and new name of the network interface when renamed
    """
    topic = SystemInfoTopic

    rhel7_name = fields.String()
    rhel8_name = fields.String()


class RenamedInterfaces(Model):
    """
    Provide list of renamed network interfaces

    These interfaces will use different names on the target system
    in comparison with original names.
    """
    topic = SystemInfoTopic

    renamed = fields.List(fields.Model(RenamedInterface))



import fnmatch
import os

from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import SystemdServiceFile, SystemdServicePreset

SYSTEMD_SYMLINKS_DIR = '/etc/systemd/system/'

_SYSTEMCTL_CMD_OPTIONS = ['--type=service', '--all', '--plain', '--no-legend']
_USR_PRESETS_PATH = '/usr/lib/systemd/system-preset/'
_ETC_PRESETS_PATH = '/etc/systemd/system-preset/'

SYSTEMD_SYSTEM_LOAD_PATH = [
    '/etc/systemd/system',
    '/usr/lib/systemd/system'
]


def get_broken_symlinks():
    """
    Get broken systemd symlinks on the system

    :return: List of broken systemd symlinks
    :rtype: list[str]
    :raises: CalledProcessError: if the `find` command fails
    :raises: OSError: if the find utility is not found
    """
    try:
        return run(['find', SYSTEMD_SYMLINKS_DIR, '-xtype', 'l'], split=True)['stdout']
    except (OSError, CalledProcessError):
        api.current_logger().error('Cannot obtain the list of broken systemd symlinks.')
        raise


def _try_call_unit_command(command, unit):
    try:
        # it is possible to call this on multiple units at once,
        # but failing to enable one service would cause others to not enable as well
        run(['systemctl', command, unit])
    except CalledProcessError as err:
        msg = 'Failed to {} systemd unit "{}". Message: {}'.format(command, unit, str(err))
        api.current_logger().error(msg)
        raise err


def enable_unit(unit):
    """
    Enable a systemd unit

    It is strongly recommended to produce SystemdServicesTasks message instead,
    unless it is absolutely necessary to handle failure yourself.

    :param unit: The systemd unit to enable
    :raises CalledProcessError: In case of failure
    """
    _try_call_unit_command('enable', unit)


def disable_unit(unit):
    """
    Disable a systemd unit

    It is strongly recommended to produce SystemdServicesTasks message instead,
    unless it is absolutely necessary to handle failure yourself.

    :param unit: The systemd unit to disable
    :raises CalledProcessError: In case of failure
    """
    _try_call_unit_command('disable', unit)


def reenable_unit(unit):
    """
    Re-enable a systemd unit

    It is strongly recommended to produce SystemdServicesTasks message, unless it
    is absolutely necessary to handle failure yourself.

    :param unit: The systemd unit to re-enable
    :raises CalledProcessError: In case of failure
    """
    _try_call_unit_command('reenable', unit)


def get_service_files():
    """
    Get list of unit files of systemd services on the system

    The list includes template units.

    :return: List of service unit files with states
    :rtype: list[SystemdServiceFile]
    :raises: CalledProcessError: in case of failure of `systemctl` command
    """
    services_files = []
    try:
        cmd = ['systemctl', 'list-unit-files'] + _SYSTEMCTL_CMD_OPTIONS
        service_units_data = run(cmd, split=True)['stdout']
    except CalledProcessError as err:
        api.current_logger().error('Cannot obtain the list of unit files:{}'.format(str(err)))
        raise

    for entry in service_units_data:
        columns = entry.split()
        services_files.append(SystemdServiceFile(name=columns[0], state=columns[1]))
    return services_files


def _join_presets_resolving_overrides(etc_files, usr_files):
    """
    Join presets and resolve preset file overrides

    Preset files in /etc/ override those with the same name in /usr/.
    If such a file is a symlink to /dev/null, it disables the one in /usr/ instead.

    :param etc_files: Systemd preset files in /etc/
    :param usr_files: Systemd preset files in /usr/
    :return: List of preset files in /etc/ and /usr/ with overridden files removed
    """
    for etc_file in etc_files:
        filename = os.path.basename(etc_file)
        for usr_file in usr_files:
            if filename == os.path.basename(usr_file):
                usr_files.remove(usr_file)
                if os.path.islink(etc_file) and os.readlink(etc_file) == '/dev/null':
                    etc_files.remove(etc_file)

    return etc_files + usr_files


def _search_preset_files(path):
    """
    Search preset files in the given path

    Presets are search recursively in the given directory.
    If path isn't an existing directory, return empty list.

    :param path: The path to search preset files in
    :return: List of found preset files
    :rtype: list[str]
    :raises: CalledProcessError: if the `find` command fails
    :raises: OSError: if the find utility is not found
    """
    if os.path.isdir(path):
        try:
            return run(['find', path, '-name', '*.preset'], split=True)['stdout']
        except (OSError, CalledProcessError) as err:
            api.current_logger().error('Cannot obtain list of systemd preset files in {}:{}'.format(path, str(err)))
            raise
    else:
        return []


def _get_system_preset_files():
    """
    Get systemd system preset files and remove overriding entries. Entries in /run/systemd/system are ignored.

    :return: List of system systemd preset files
    :raises: CalledProcessError: if the `find` command fails
    :raises: OSError: if the find utility is not found
    """
    etc_files = _search_preset_files(_ETC_PRESETS_PATH)
    usr_files = _search_preset_files(_USR_PRESETS_PATH)

    preset_files = _join_presets_resolving_overrides(etc_files, usr_files)
    preset_files.sort()
    return preset_files


def _recursive_glob(pattern, root_dir):
    for _, _, filenames in os.walk(root_dir):
        for filename in filenames:
            if fnmatch.fnmatch(filename, pattern):
                yield filename


def _parse_preset_entry(entry, presets, load_path):
    """
    Parse a single entry (line) in a preset file

    Single entry might set presets on multiple units using globs.

    :param entry: The entry to parse
    :param presets: Dictionary to store the presets into
    :param load_path: List of paths to look systemd unit files up in
    """

    columns = entry.split()
    if len(columns) < 2 or columns[0] not in ('enable', 'disable'):
        raise ValueError('Invalid preset file entry: "{}"'.format(entry))

    for path in load_path:
        # TODO(mmatuska): This currently also globs non unit files,
        # so the results need to be filtered with something like endswith('.<unit_type>')
        unit_files = _recursive_glob(columns[1], root_dir=path)

        for unit_file in unit_files:
            if '@' in columns[1] and len(columns) > 2:
                # unit is a template,
                # if the entry contains instance names after template unit name
                # the entry only applies to the specified instances, not to the
                # template itself
                for instance in columns[2:]:
                    service_name = unit_file[:unit_file.index('@') + 1] + instance + '.service'
                    if service_name not in presets:  # first occurrence has priority
                        presets[service_name] = columns[0]

            elif unit_file not in presets:  # first occurrence has priority
                presets[unit_file] = columns[0]


def _parse_preset_files(preset_files, load_path, ignore_invalid_entries):
    """
    Parse presets from preset files

    :param load_path: List of paths to search units at
    :param ignore_invalid_entries: Whether to ignore invalid entries in preset files or raise an error
    :return: Dictionary mapping systemd units to their preset state
    :rtype: dict[str, str]
    :raises: ValueError: when a preset file has invalid content
    """
    presets = {}

    for preset in preset_files:
        with open(preset, 'r') as preset_file:
            for line in preset_file:
                stripped = line.strip()
                if stripped and stripped[0] not in ('#', ';'):  # ignore comments
                    try:
                        _parse_preset_entry(stripped, presets, load_path)
                    except ValueError as err:
                        new_msg = 'Invalid preset file {pfile}: {error}'.format(pfile=preset, error=str(err))
                        if ignore_invalid_entries:
                            api.current_logger().warning(new_msg)
                            continue
                        raise ValueError(new_msg)
    return presets


def get_system_service_preset_files(service_files, ignore_invalid_entries=False):
    """
    Get system preset files for services

    Presets for static and transient services are filtered out.

    :param services_files: List of service unit files
    :param ignore_invalid_entries: Ignore invalid entries in preset files if True, raise ValueError otherwise
    :return: List of system systemd services presets
    :rtype: list[SystemdServicePreset]
    :raises: CalledProcessError: In case of errors when discovering systemd preset files
    :raises: OSError: When the `find` command is not available
    :raises: ValueError: When a preset file has invalid content and ignore_invalid_entries is False
    """
    preset_files = _get_system_preset_files()
    presets = _parse_preset_files(preset_files, SYSTEMD_SYSTEM_LOAD_PATH, ignore_invalid_entries)

    preset_models = []
    for unit, state in presets.items():
        if unit.endswith('.service'):
            service_file = next(iter([s for s in service_files if s.name == unit]), None)
            # presets can also be set on instances of template services which don't have a unit file
            if service_file and service_file.state in ('static', 'transient'):
                continue
            preset_models.append(SystemdServicePreset(service=unit, state=state))

    return preset_models



import errno
import os
import re

from leapp.libraries.common.config import get_env
from leapp.libraries.stdlib import api
from leapp.models import (
    InitrdIncludes,
    PersistentNetNamesFacts,
    PersistentNetNamesFactsInitramfs,
    RenamedInterface,
    RenamedInterfaces,
    TargetInitramfsTasks
)
from leapp.utils.deprecation import suppress_deprecation

LINK_FILE_TEMPLATE = """# Generated by LEAPP
[Match]
MACAddress={}

[Link]
Name={}
"""


def generate_link_file(interface):
    try:
        os.makedirs('/etc/systemd/network')
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    link_file = '/etc/systemd/network/10-leapp-{}.link'.format(interface.name)
    with open(link_file, 'w') as f:
        f.write(LINK_FILE_TEMPLATE.format(interface.mac, interface.name))

    return link_file


@suppress_deprecation(InitrdIncludes)
def process():

    if get_env('LEAPP_NO_NETWORK_RENAMING', '0') == '1':
        api.current_logger().info(
            'Skipping handling of possibly renamed network interfaces: leapp executed with LEAPP_NO_NETWORK_RENAMING=1'
        )
        return

    rhel7_ifaces = next(api.consume(PersistentNetNamesFacts)).interfaces
    rhel8_ifaces = next(api.consume(PersistentNetNamesFactsInitramfs)).interfaces

    rhel7_ifaces_map = {iface.mac: iface for iface in rhel7_ifaces}
    rhel8_ifaces_map = {iface.mac: iface for iface in rhel8_ifaces}

    initrd_files = []
    missing_ifaces = []
    renamed_interfaces = []

    if rhel7_ifaces != rhel8_ifaces:
        for iface in rhel7_ifaces:
            rhel7_name = rhel7_ifaces_map[iface.mac].name
            try:
                rhel8_name = rhel8_ifaces_map[iface.mac].name
            except KeyError:
                missing_ifaces.append(iface)
                api.current_logger().warning(
                    'The device with MAC "{}" is not detected in the upgrade'
                    ' environment. Required driver: "{}".'
                    ' Original interface name: "{}".'
                    .format(iface.mac, iface.driver, iface.name)
                )
                continue

            if rhel7_name != rhel8_name and get_env('LEAPP_NO_NETWORK_RENAMING', '0') != '1':
                api.current_logger().warning('Detected interface rename {} -> {}.'.format(rhel7_name, rhel8_name))

                if re.search('eth[0-9]+', iface.name) is not None:
                    api.current_logger().warning('Interface named using eth prefix, refusing to generate link file')
                    renamed_interfaces.append(RenamedInterface(**{'rhel7_name': rhel7_name,
                                                                  'rhel8_name': rhel8_name}))
                    continue

                initrd_files.append(generate_link_file(iface))

    if missing_ifaces:
        msg = (
            'Some network devices have not been detected inside the'
            ' upgrade environment and so related network interfaces'
            ' could be renamed on the upgraded system.'
        )
        # Note(pstodulk):
        # This usually happens when required (RHEL 8 compatible)
        # drivers are not included in the upgrade initramfs.
        # We can add more information later. Currently we cannot provide
        # better instructions for users before (at least):
        # a) networking work in the upgrade initramfs (PR #583)
        # b) it's possible to influence the upgrade initramfs (PR #517)
        # TODO(pstodulk): gen report msg
        api.current_logger().warning(msg)

    api.produce(RenamedInterfaces(renamed=renamed_interfaces))
    api.produce(InitrdIncludes(files=initrd_files))
    # TODO: cover actor by tests in future. I am skipping writing of tests
    # now as some refactoring and bugfixing related to this actor
    # is planned already.
    api.produce(TargetInitramfsTasks(include_files=initrd_files))



from leapp.actors import Actor
from leapp.libraries.actor import opensshsubsystemsftp
from leapp.models import DistributionSignedRPM, OpenSshConfig
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class OpenSshSubsystemSftp(Actor):
    """
    The RHEL9 changes the SCP to use SFTP protocol internally. The both RHEL8 and RHEL9
    enable SFTP server by default, but if the user disabled the SFTP for some reason,
    it might make sense to warn that some previously working SCP operations could stop
    working.
    """

    name = 'open_ssh_subsystem_sftp'
    consumes = (OpenSshConfig, DistributionSignedRPM,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        opensshsubsystemsftp.process(self.consume(OpenSshConfig))



from leapp import reporting
from leapp.actors import Actor
from leapp.models import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class UnsupportedUpgradeCheck(Actor):
    """
    Checks environment variables and produces a warning report if the upgrade is unsupported.

    Upgrade is unsupported if any LEAPP_DEVEL_* variable is used or an experimental actor is enabled.
    This can be overridden by setting the variable LEAPP_UNSUPPORTED (at user's own risk).
    """

    name = 'unsupported_upgrade_check'
    consumes = ()
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        leapp_vars = self.configuration.leapp_env_vars
        devel_vars = [v for v in leapp_vars if v.name.startswith('LEAPP_DEVEL_')]
        experimental = bool([v for v in leapp_vars if v.name == 'LEAPP_EXPERIMENTAL' and v.value == '1'])
        override = bool([v for v in leapp_vars if v.name == 'LEAPP_UNSUPPORTED' and v.value == '1'])

        if override:
            reporting.create_report([
                reporting.Title('Upgrade is unsupported'),
                reporting.Summary(
                    'Environment variable LEAPP_UNSUPPORTED has been detected. A successful and safe '
                    'upgrade process cannot be guaranteed. From now on you are continuing at your own '
                    'risk.\n'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups([reporting.Groups.UPGRADE_PROCESS, reporting.Groups.SANITY]),
            ])

        else:
            if devel_vars:
                reporting.create_report([
                    reporting.Title('Upgrade inhibited due to usage of development variables'),
                    reporting.Summary(
                        'One or more environment variables in the form of LEAPP_DEVEL_* have been detected.\n'
                        'These variables are for development purposes only and can interfere with the upgrade '
                        'process in unexpected ways. As such, a successful and safe upgrade process cannot be '
                        'guaranteed and the upgrade is unsupported.\n'
                        'You can bypass this error by setting the LEAPP_UNSUPPORTED variable but by doing so, '
                        'you continue at your own risk.\n'
                        'Found development variables:\n- {}\n'.format('\n- '.join([v.name for v in devel_vars]))
                    ),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.INHIBITOR]),
                    reporting.Groups([reporting.Groups.UPGRADE_PROCESS, reporting.Groups.SANITY]),
                    reporting.Remediation(hint=('Invoke leapp without any LEAPP_DEVEL_* environment variables '
                                                'or set LEAPP_UNSUPPORTED=1.'))
                ])

            if experimental:
                reporting.create_report([
                    reporting.Title('Upgrade inhibited due to enabled experimental actors'),
                    reporting.Summary(
                        'One or more enabled experimental actors have been detected.\n'
                        'These actors are unstable or in development and can interfere with the upgrade '
                        'process in unexpected ways. As such, a successful and safe upgrade process cannot be '
                        'guaranteed and the upgrade is unsupported.\n'
                        'You can bypass this error by setting the LEAPP_UNSUPPORTED variable but by doing so, '
                        'you continue at your own risk.\n'
                    ),
                    reporting.Severity(reporting.Severity.HIGH),
                    reporting.Groups([reporting.Groups.INHIBITOR]),
                    reporting.Groups([reporting.Groups.UPGRADE_PROCESS, reporting.Groups.SANITY]),
                    reporting.Remediation(hint=('Invoke leapp without any --whitelist-experimental options '
                                                'or set LEAPP_UNSUPPORTED=1.'))
                ])



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class CustomModifications(Model):
    """Model to store any custom or modified files that are discovered in leapp directories"""
    topic = SystemFactsTopic

    filename = fields.String()
    actor_name = fields.String()
    type = fields.StringEnum(choices=['custom', 'modified'])
    rpm_checks_str = fields.String(default='')
    component = fields.String()



from leapp.actors import Actor
from leapp.libraries.actor import modscan
from leapp.tags import FactsPhaseTag, IPUWorkflowTag
from leapp.utils.deprecation import suppress_deprecation

from leapp.models import (  # isort:skip
    RequiredUpgradeInitramPackages,  # deprecated
    UpgradeDracutModule,  # deprecated
    TargetUserSpaceUpgradeTasks,
    UpgradeInitramfsTasks
)


@suppress_deprecation(RequiredUpgradeInitramPackages, UpgradeDracutModule)
class CommonLeappDracutModules(Actor):
    """
    Influences the generation of the initram disk

    The initram disk generation is influenced by specifying necessary dracut modules and packages that are
    required to be installed in the target userspace so required fields can be included.
    Modules to be added are specified via the UpgradeDracutModule message.
    Packages to install on the target userspace are specified by the RequiredUpgradeInitramPackages message.
    """

    name = 'common_leapp_dracut_modules'
    consumes = ()
    produces = (
        RequiredUpgradeInitramPackages,  # deprecated
        TargetUserSpaceUpgradeTasks,
        UpgradeDracutModule,  # deprecated
        UpgradeInitramfsTasks,
    )
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        modscan.process()



from leapp.actors import Actor
from leapp.libraries.actor import vsftpdconfigread
from leapp.models import DistributionSignedRPM, VsftpdFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class VsftpdConfigRead(Actor):
    """
    Reads vsftpd configuration files (/etc/vsftpd/*.conf) and extracts necessary information.
    """

    name = 'vsftpd_config_read'
    consumes = (DistributionSignedRPM,)
    produces = (VsftpdFacts,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        installed_rpm_facts = next(self.consume(DistributionSignedRPM))
        if vsftpdconfigread.is_processable(installed_rpm_facts):
            self.produce(vsftpdconfigread.get_vsftpd_facts())



from leapp.actors import Actor
from leapp.libraries.actor import multipathconfread
from leapp.models import DistributionSignedRPM, MultipathConfFacts8to9, TargetUserSpaceUpgradeTasks
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class MultipathConfRead8to9(Actor):
    """
    Read multipath configuration files and extract the necessary information

    Related files:
      - /etc/multipath.conf
      - /etc/multipath/ - any files inside the directory
      - /etc/xdrdevices.conf

    As well, create task (msg) to copy all needed multipath files into
    the target container as the files are needed to create proper initramfs.
    This covers the files mentioned above.
    """

    name = 'multipath_conf_read_8to9'
    consumes = (DistributionSignedRPM,)
    produces = (MultipathConfFacts8to9, TargetUserSpaceUpgradeTasks)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        if multipathconfread.is_processable():
            res = multipathconfread.get_multipath_conf_facts()
            if res:
                self.produce(res)
                # Create task to copy multipath config files Iff facts
                # are generated
                multipathconfread.produce_copy_to_target_task()



import errno
import textwrap
from os.path import basename

import mock
import six

from leapp.libraries.actor import ifcfgscanner
from leapp.libraries.common.testutils import make_OSError, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import IfCfg

_builtins_open = "builtins.open" if six.PY3 else "__builtin__.open"


def _listdir_ifcfg(path):
    if path == ifcfgscanner.SYSCONFIG_DIR:
        return ["ifcfg-net0"]
    raise make_OSError(errno.ENOENT)


def _listdir_ifcfg2(path):
    if path == ifcfgscanner.SYSCONFIG_DIR:
        return ["ifcfg-net0", "ifcfg-net1"]
    raise make_OSError(errno.ENOENT)


def _exists_ifcfg(filename):
    return basename(filename).startswith("ifcfg-")


def _exists_keys(filename):
    if _exists_ifcfg(filename):
        return True
    return basename(filename).startswith("keys-")


def test_no_conf(monkeypatch):
    """
    No report if there are no ifcfg files.
    """

    monkeypatch.setattr(ifcfgscanner, "listdir", lambda _: ())
    monkeypatch.setattr(api, "produce", produce_mocked())
    ifcfgscanner.process()
    assert not api.produce.called


def test_ifcfg1(monkeypatch):
    """
    Parse a single ifcfg file.
    """

    ifcfg_file = textwrap.dedent("""
        TYPE=Wireless  # Some comment
        # Another comment
        ESSID=wep1
        NAME="wep1"
        MODE='Managed'   # comment
        WEP_KEY_FLAGS=ask
        SECURITYMODE=open
        DEFAULTKEY=1
        KEY_TYPE=key
    """)

    mock_config = mock.mock_open(read_data=ifcfg_file)
    with mock.patch(_builtins_open, mock_config):
        monkeypatch.setattr(ifcfgscanner, "listdir", _listdir_ifcfg)
        monkeypatch.setattr(ifcfgscanner.path, "exists", _exists_ifcfg)
        monkeypatch.setattr(api, "produce", produce_mocked())
        ifcfgscanner.process()

        assert api.produce.called == 1
        assert len(api.produce.model_instances) == 1
        ifcfg = api.produce.model_instances[0]
        assert isinstance(ifcfg, IfCfg)
        assert ifcfg.filename == "/etc/sysconfig/network-scripts/ifcfg-net0"
        assert ifcfg.secrets is None
        assert len(ifcfg.properties) == 8
        assert ifcfg.properties[0].name == "TYPE"
        assert ifcfg.properties[0].value == "Wireless"
        assert ifcfg.properties[1].name == "ESSID"
        assert ifcfg.properties[1].value == "wep1"
        assert ifcfg.properties[2].name == "NAME"
        assert ifcfg.properties[2].value == "wep1"
        assert ifcfg.properties[3].name == "MODE"
        assert ifcfg.properties[3].value == "Managed"


def test_ifcfg2(monkeypatch):
    """
    Parse two ifcfg files.
    """

    mock_config = mock.mock_open(read_data="TYPE=Ethernet")
    with mock.patch(_builtins_open, mock_config):
        monkeypatch.setattr(ifcfgscanner, "listdir", _listdir_ifcfg2)
        monkeypatch.setattr(ifcfgscanner.path, "exists", _exists_ifcfg)
        monkeypatch.setattr(api, "produce", produce_mocked())
        ifcfgscanner.process()

        assert api.produce.called == 2
        assert len(api.produce.model_instances) == 2
        ifcfg = api.produce.model_instances[0]
        assert isinstance(ifcfg, IfCfg)


def test_ifcfg_key(monkeypatch):
    """
    Report ifcfg secrets from keys- file.
    """

    mock_config = mock.mock_open(read_data="KEY_PASSPHRASE1=Hell0")
    with mock.patch(_builtins_open, mock_config):
        monkeypatch.setattr(ifcfgscanner, "listdir", _listdir_ifcfg)
        monkeypatch.setattr(ifcfgscanner.path, "exists", _exists_keys)
        monkeypatch.setattr(api, "produce", produce_mocked())
        ifcfgscanner.process()

        assert api.produce.called == 1
        assert len(api.produce.model_instances) == 1
        ifcfg = api.produce.model_instances[0]
        assert isinstance(ifcfg, IfCfg)
        assert ifcfg.filename == "/etc/sysconfig/network-scripts/ifcfg-net0"
        assert len(ifcfg.secrets) == 1
        assert ifcfg.secrets[0].name == "KEY_PASSPHRASE1"
        assert ifcfg.secrets[0].value is None



from leapp.libraries.actor.readopensshconfig import line_empty, parse_config, produce_config
from leapp.models import OpenSshConfig, OpenSshPermitRootLogin


def test_line_empty():
    assert line_empty("".strip()) is True
    assert line_empty("     ".strip()) is True
    assert line_empty("   # comment".strip()) is True
    assert line_empty("# comment".strip()) is True
    assert line_empty("option".strip()) is False
    assert line_empty("    option".strip()) is False


def test_parse_config():
    config = [
        "# comment from file",
        "",  # empty line
        "   ",  # whitespace line
        "permitrootlogin prohibit-password",
        "permittty yes",
        "useprivilegeseparation no",
        "protocol 2",
        "hostkey /etc/ssh/ssh_host_ecdsa_key",  # unrelated duplicate keys
        "hostkey /etc/ssh/ssh_host_ed25519_key",
        "ciphers aes128-ctr",
        "macs hmac-md5",
        "subsystem sftp internal-sftp",
        "subsystem other internal-other",  # this is ignored
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert len(output.permit_root_login) == 1
    assert output.permit_root_login[0].value == "prohibit-password"
    assert output.use_privilege_separation == "no"
    assert output.protocol == "2"
    assert output.ciphers == "aes128-ctr"
    assert output.macs == "hmac-md5"
    assert output.subsystem_sftp == "internal-sftp"


def test_parse_config_case():
    config = [
        "PermitRootLogin prohibit-password",
        "UsePrivilegeSeparation yes",
        "Protocol 1",
        "SubSystem sftp sftp-server",
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert len(output.permit_root_login) == 1
    assert output.permit_root_login[0].value == "prohibit-password"
    assert output.use_privilege_separation == "yes"
    assert output.protocol == "1"
    assert output.subsystem_sftp == "sftp-server"


def test_parse_config_multiple():
    config = [
        "PermitRootLogin prohibit-password",
        "PermitRootLogin no",
        "PermitRootLogin yes",
        "Ciphers aes128-cbc",
        "Ciphers aes256-cbc",
        "subsystem sftp internal-sftp",
        "subsystem sftp internal-sftp2",
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert len(output.permit_root_login) == 3
    assert output.permit_root_login[0].value == "prohibit-password"
    assert output.permit_root_login[1].value == "no"
    assert output.permit_root_login[2].value == "yes"
    assert output.use_privilege_separation is None
    assert output.protocol is None
    assert output.ciphers == 'aes128-cbc'
    assert output.subsystem_sftp == 'internal-sftp'


def test_parse_config_commented():
    config = [
        "#PermitRootLogin no",
        "#UsePrivilegeSeparation no",
        "#Protocol 12",
        "#SubSystem sftp internal-sftp",
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert not output.permit_root_login
    assert output.use_privilege_separation is None
    assert output.protocol is None
    assert output.subsystem_sftp is None


def test_parse_config_missing_argument():
    config = [
        "PermitRootLogin",
        "UsePrivilegeSeparation",
        "Protocol"
        "SubSystem"
        "SubSystem sftp"
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert not output.permit_root_login
    assert output.use_privilege_separation is None
    assert output.protocol is None
    assert output.subsystem_sftp is None


def test_parse_config_match():
    config = [
        "PermitRootLogin yes",
        "Match address 192.168.*",
        "   PermitRootLogin no"
    ]

    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert len(output.permit_root_login) == 2
    assert output.permit_root_login[0].value == 'yes'
    assert output.permit_root_login[0].in_match is None
    assert output.permit_root_login[1].value == 'no'
    assert output.permit_root_login[1].in_match == ['address', '192.168.*']
    assert output.use_privilege_separation is None
    assert output.protocol is None


def test_parse_config_deprecated():
    config = [
        "permitrootlogin without-password"
    ]
    output = parse_config(config)
    assert isinstance(output, OpenSshConfig)
    assert len(output.permit_root_login) == 1
    assert output.permit_root_login[0].value == "prohibit-password"


def test_parse_config_empty():
    output = parse_config([])
    assert isinstance(output, OpenSshConfig)
    assert isinstance(output, OpenSshConfig)
    assert not output.permit_root_login
    assert output.use_privilege_separation is None
    assert output.protocol is None


def test_produce_config():
    output = []

    def fake_producer(*args):
        output.extend(args)

    config = OpenSshConfig(
        permit_root_login=[OpenSshPermitRootLogin(value="no")],
        use_privilege_separation="yes",
        protocol="1",
        deprecated_directives=[],
        subsystem_sftp="internal-sftp",
    )

    produce_config(fake_producer, config)
    assert len(output) == 1
    cfg = output[0]
    assert len(cfg.permit_root_login) == 1
    assert cfg.permit_root_login[0].value == "no"
    assert cfg.use_privilege_separation == "yes"
    assert cfg.protocol == '1'
    assert cfg.subsystem_sftp == 'internal-sftp'


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(OpenSshConfig)



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api

# The devel variable for target product channel can also contain 'beta'
SUPPORTED_TARGET_CHANNELS = {'ga', 'e4s', 'eus', 'aus'}
CONSUMED_DATA_STREAM_ID = '3.0'


def get_env(name, default=None):
    """Return Leapp environment variable value if matched by name."""
    for var in api.current_actor().configuration.leapp_env_vars:
        if var.name == name:
            return var.value
    return default


def get_all_envs():
    """Return all Leapp environment variables (both name and value)."""
    return api.current_actor().configuration.leapp_env_vars


def get_product_type(sys_type):
    """
    Get expected product type (ga/beta/htb) of the chosen sys type.

    By default, expected product type is 'ga'. It can be changed using these envars:
        LEAPP_DEVEL_SOURCE_PRODUCT_TYPE
        LEAPP_DEVEL_TARGET_PRODUCT_TYPE
    which can set the product type of chosen system (source/target) to one of
    valid product types: ga/beta/htb.

    Raise ValueError when specified sys_type or set product_type is invalid.

    :param sys_type: choose system for which to get the product type: 'source' or 'target'
    :type sys_type: str
    :return: 'ga' (default), 'htb', 'beta'
    :rtype: str
    """
    if sys_type == 'source':
        envar = 'LEAPP_DEVEL_SOURCE_PRODUCT_TYPE'
    elif sys_type == 'target':
        envar = 'LEAPP_DEVEL_TARGET_PRODUCT_TYPE'
    else:
        raise ValueError('Given invalid sys_type. Valid values: source/target')
    val = get_env(envar, '').lower()
    if not val:
        return 'ga'
    if val in ('beta', 'htb', 'ga'):
        return val
    raise ValueError('Invalid value in the {} envar. Possible values: ga/beta/htb.'.format(envar))


def _stop_exection_due_invalid_channel(channel_specified_via, actual_value, supported_values):
    raise StopActorExecutionError('Target channel specified via {0} has invalid value: {1}. Possible values: {2}'
                                  .format(channel_specified_via, actual_value, ", ".join(supported_values)))


def get_target_product_channel(default='ga'):
    """
    Get target product channel specified when running the IPU or default if no channel was specified.
    The channel returned is always lower case regardless how it was specified by the user.

    The target channel can be specified via:
        - Using the environment variable LEAPP_DEVEL_TARGET_PRODUCT_TYPE (devel variable with higher priority than
        any other way of specifying target channel).
        - Using the environment variable LEAPP_TARGET_PRODUCT_CHANNEL
        - Using the '--channel' option when running leapp preupgrade/upgrade

    :param default: Value to be returned if no target product type has been specified when running leapp.
    :type default: str
    :returns: The user-specified target channel or default if no channel was specified.
    :rtype: str
    """

    devel_target_product_type = get_env('LEAPP_DEVEL_TARGET_PRODUCT_TYPE')
    if devel_target_product_type:
        devel_target_product_type = devel_target_product_type.lower()
        supported_devel_target_channels = SUPPORTED_TARGET_CHANNELS | {'beta'}
        if devel_target_product_type not in supported_devel_target_channels:
            _stop_exection_due_invalid_channel('LEAPP_DEVEL_TARGET_PRODUCT_TYPE',
                                               devel_target_product_type,
                                               supported_devel_target_channels)
        return devel_target_product_type

    target_product_channel = get_env('LEAPP_TARGET_PRODUCT_CHANNEL')
    if target_product_channel:
        target_product_channel = target_product_channel.lower()
        if target_product_channel not in SUPPORTED_TARGET_CHANNELS:
            _stop_exection_due_invalid_channel('LEAPP_TARGET_PRODUCT_CHANNEL or the --channel option',
                                               target_product_channel,
                                               SUPPORTED_TARGET_CHANNELS)
        return target_product_channel

    return default


def get_consumed_data_stream_id():
    """Get the identifier of the asset family used by leapp."""
    return CONSUMED_DATA_STREAM_ID



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class NetworkManagerConnectionProperty(Model):
    """
    Name-value pair for NetworkManager properties.

    This model is not expected to be used as a message (produced/consumed by actors).
    It is used within NetworkManagerConnectionSetting of a NetworkManagerConnection.
    """
    topic = SystemInfoTopic

    name = fields.String()
    """ Name of a property """
    value = fields.String()
    """ Value of a property """


class NetworkManagerConnectionSetting(Model):
    """
    NetworkManager setting, composed of a name and a list of name-value pairs.

    This model is not expected to be used as a message (produced/consumed by actors).
    It is used within NetworkManagerConnection.
    """
    topic = SystemInfoTopic

    name = fields.String()
    """ The NetworkManager setting name """
    properties = fields.List(fields.Model(NetworkManagerConnectionProperty), default=[])
    """ The name-value pair for every setting property """


class NetworkManagerConnection(Model):
    """
    NetworkManager native keyfile connection

    Produced for every connection profile loaded from INI-stile files
    described in nm-settings-keyfile(5) manual.
    """
    topic = SystemInfoTopic

    settings = fields.List(fields.Model(NetworkManagerConnectionSetting), default=[])
    """ List of NetworkManager settings """
    filename = fields.String()
    """ Path to file this model was populated from """



from leapp.actors import Actor
from leapp.libraries.actor.xorgdriverlib import check_drv_and_options, get_xorg_logs_from_journal
from leapp.libraries.stdlib import api
from leapp.models import XorgDrvFacts
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class XorgDrvFacts8to9(Actor):
    """
    Check the journal logs for deprecated Xorg drivers.

    This actor checks the journal logs and looks for deprecated Xorg drivers.
    """

    name = 'xorgdrvfacts8to9'
    consumes = ()
    produces = (XorgDrvFacts,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        xorg_logs = get_xorg_logs_from_journal()
        deprecated_drivers = []
        for driver in ['RADEON', 'ATI', 'AMDGPU', 'MACH64', 'intel', 'spiceqxl', 'qxl', 'NOUVEAU', 'NV', 'VESA']:
            deprecated_driver = check_drv_and_options(driver, xorg_logs)
            if deprecated_driver:
                deprecated_drivers.append(deprecated_driver)

        api.produce(XorgDrvFacts(xorg_drivers=deprecated_drivers))



import os

import pytest

from leapp.libraries.actor import copydnfconfintotargetuserspace
from leapp.libraries.common.testutils import logger_mocked, produce_mocked


@pytest.mark.parametrize(
    "userspace_conf_exists,expected",
    [(False, "/etc/dnf/dnf.conf"), (True, "/etc/leapp/files/dnf.conf")],
)
def test_copy_correct_dnf_conf(monkeypatch, userspace_conf_exists, expected):
    monkeypatch.setattr(os.path, "exists", lambda _: userspace_conf_exists)

    mocked_produce = produce_mocked()
    monkeypatch.setattr(copydnfconfintotargetuserspace.api, 'produce', mocked_produce)
    monkeypatch.setattr(copydnfconfintotargetuserspace.api, 'current_logger', logger_mocked())

    copydnfconfintotargetuserspace.process()

    assert mocked_produce.called == 1
    assert len(mocked_produce.model_instances) == 1
    assert len(mocked_produce.model_instances[0].copy_files) == 1
    assert mocked_produce.model_instances[0].copy_files[0].src == expected
    assert mocked_produce.model_instances[0].copy_files[0].dst == "/etc/dnf/dnf.conf"



import errno
import re

import leapp.libraries.common.spamassassinutils as utils
from leapp.libraries.stdlib import api


def _drop_ssl_version(value):
    regex = r'(?<![\w-])--ssl-version(=|\s+)(sslv3|tlsv1)(?![\w-])'
    if not re.search(r'(?<![\w-])--ssl(?![\w-])', value):
        value = re.sub(regex, '--ssl', value, count=1)
    return re.sub(regex, '', value)


def _drop_daemonize_option(value):
    shortopts = utils.SPAMD_SHORTOPTS_NOARG
    value = re.sub(r'(?<![\w-])-d+(?![\w-])',
                   r'', value)
    while True:
        value, nsubs = re.subn(r'(?<![\w-])(-[' + shortopts + ']*)d',
                               r'\1', value)
        if nsubs == 0:
            break
    return re.sub(r'(?<![\w-])--daemonize(?![\w-])', '', value)


def _rewrite_spamd_option(content, ops):
    pre_assignment, assignment, post_assignment = \
        utils.parse_sysconfig_spamassassin(content)
    if assignment:
        value = re.sub(r'^\s*' + utils.SYSCONFIG_VARIABLE + r'\s*=(.*)$',
                       r'\1', assignment)
        for op in ops:
            value = op(value)
        assignment = utils.SYSCONFIG_VARIABLE + '=' + value
    res = '\n'.join(filter(None, (pre_assignment, assignment, post_assignment)))
    if res and not res.endswith('\n'):
        res += '\n'
    return res


def _rewrite_spamd_config(facts, content):
    ops = []
    if facts.spamd_ssl_version:
        ops.append(_drop_ssl_version)
    if not facts.service_overriden:
        ops.append(_drop_daemonize_option)
    return _rewrite_spamd_option(content, ops) if ops else content


def migrate_spamd_config(facts, fileops, backup_func):
    """
    Removes --ssl-version and -d/--daemonize options from the spamassassin
    sysconfig file. The file is backed up beforehand.
    """
    nothing_to_migrate_msg = 'There is nothing to migrate in the spamd configuration file'
    if facts.service_overriden and not facts.spamd_ssl_version:
        api.current_logger().info(nothing_to_migrate_msg)
        return
    try:
        content = fileops.read(utils.SYSCONFIG_SPAMASSASSIN)
    except (OSError, IOError) as e:
        if e.errno == errno.ENOENT:
            api.current_logger().info(nothing_to_migrate_msg)
        else:
            api.current_logger().warning('Failed to read spamd configuration file: %s' % e)
        return
    new_content = _rewrite_spamd_config(facts, content)
    if new_content == content:
        api.current_logger().info('There is nothing to migrate in the spamd configuration file')
        return
    try:
        backup_path = backup_func(utils.SYSCONFIG_SPAMASSASSIN)
        api.current_logger().info('spamd configuration file backup created at %s.'
                                  % backup_path)
    except (OSError, IOError) as e:
        api.current_logger().warning(
            'spamd configuration file migration will not be performed. Failed to create backup: %s' % e)
        return
    try:
        fileops.write(utils.SYSCONFIG_SPAMASSASSIN, new_content)
    except (OSError, IOError) as e:
        api.current_logger().warning('Failed to rewrite the spamd configuration file: %s' % e)



import os

from leapp.libraries.actor import xorgdriverlib
from leapp.models import XorgDrv, XorgDrvFacts

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def _read_log_file(path):
    """
    Read a log file in text mode and return the contents as an array.

    :param path: Log file path
    """
    with open(path, 'r') as f:
        return f.read().splitlines()


def test_check_drv_and_options_qxl_driver(monkeypatch):

    def get_xorg_logs_from_journal_mocked():
        return _read_log_file(os.path.join(CUR_DIR, 'files/journalctl-xorg-qxl'))

    monkeypatch.setattr(xorgdriverlib, 'get_xorg_logs_from_journal', get_xorg_logs_from_journal_mocked)
    xorg_logs = xorgdriverlib.get_xorg_logs_from_journal()
    expected = XorgDrv(driver='qxl', has_options=False)
    actual = xorgdriverlib.check_drv_and_options('qxl', xorg_logs)
    assert expected == actual


def test_check_drv_and_options_intel_driver(monkeypatch):

    def get_xorg_logs_from_journal_mocked():
        return _read_log_file(os.path.join(CUR_DIR, 'files/journalctl-xorg-intel'))

    monkeypatch.setattr(xorgdriverlib, 'get_xorg_logs_from_journal', get_xorg_logs_from_journal_mocked)
    xorg_logs = xorgdriverlib.get_xorg_logs_from_journal()
    expected = XorgDrv(driver='intel', has_options=True)
    actual = xorgdriverlib.check_drv_and_options('intel', xorg_logs)
    assert expected == actual


def test_actor_with_deprecated_driver_without_options(current_actor_context, monkeypatch):

    def get_xorg_logs_from_journal_mocked():
        return _read_log_file(os.path.join(CUR_DIR, 'files/journalctl-xorg-qxl'))

    monkeypatch.setattr(xorgdriverlib, 'get_xorg_logs_from_journal', get_xorg_logs_from_journal_mocked)
    current_actor_context.run()
    facts = list(current_actor_context.consume(XorgDrvFacts))
    assert facts and len(facts[0].xorg_drivers) == 1
    assert (facts[0].xorg_drivers)[0].driver == 'qxl'
    assert (facts[0].xorg_drivers)[0].has_options is False


def test_actor_with_deprecated_driver_with_options(current_actor_context, monkeypatch):

    def get_xorg_logs_from_journal_mocked():
        return _read_log_file(os.path.join(CUR_DIR, 'files/journalctl-xorg-intel'))

    monkeypatch.setattr(xorgdriverlib, 'get_xorg_logs_from_journal', get_xorg_logs_from_journal_mocked)
    current_actor_context.run()
    facts = list(current_actor_context.consume(XorgDrvFacts))
    assert facts and len(facts[0].xorg_drivers) == 1
    assert (facts[0].xorg_drivers)[0].driver == 'intel'
    assert (facts[0].xorg_drivers)[0].has_options is True


def test_actor_without_deprecated_driver(current_actor_context, monkeypatch):

    def get_xorg_logs_from_journal_mocked():
        return _read_log_file(os.path.join(CUR_DIR, 'files/journalctl-xorg-without-qxl'))

    monkeypatch.setattr(xorgdriverlib, 'get_xorg_logs_from_journal', get_xorg_logs_from_journal_mocked)
    current_actor_context.run()
    facts = current_actor_context.consume(XorgDrvFacts)
    assert facts and len(facts[0].xorg_drivers) == 0



import re

from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.stdlib import run

# When the output spans multiple lines, each of the lines after the first one
# start with a '    <SPACES>    : '
LOADED_PLUGINS_NEXT_LINE_START = ' +: '


def _parse_loaded_plugins(package_manager_output):
    """
    Retrieves a list of plugins that are being loaded when calling dnf/yum.

    :param dict package_manager_output: The result of running the package manager command.
    :rtype: list
    :returns: A list of plugins that are being loaded by the package manager.
    """
    # Package manager might break the information about loaded plugins into multiple lines,
    # we need to concaternate the list ourselves
    loaded_plugins_str = ''
    for line in package_manager_output['stdout']:
        if line.startswith('Loaded plugins:'):
            # We have found the first line that contains the plugins
            plugins_on_this_line = line[16:]  # Remove the `Loaded plugins: ` part

            if plugins_on_this_line[-1] == ',':
                plugins_on_this_line += ' '

            loaded_plugins_str += plugins_on_this_line
            continue

        if loaded_plugins_str:
            if re.match(LOADED_PLUGINS_NEXT_LINE_START, line):
                # The list of plugins continues on this line
                plugins_on_this_line = line.lstrip(' :')  # Remove the leading spaces and semicolon

                # Plugins are separated by ', ', however the space at the end of line might get dropped, add it
                # so we can split it by ', ' later
                if plugins_on_this_line[-1] == ',':
                    plugins_on_this_line += ' '

                loaded_plugins_str += plugins_on_this_line
            else:
                # The list of loaded plugins ended
                break

    return loaded_plugins_str.split(', ')


def scan_enabled_package_manager_plugins():
    """
    Runs package manager (yum/dnf) command and parses its output for enabled/loaded plugins.

    :return: A list of enabled plugins.
    :rtype: List
    """

    # We rely on package manager itself to report what plugins are used when it is invoked.
    # An alternative approach would be to check the install path for package manager plugins
    # and parse corresponding plugin configuration files.

    if get_source_major_version() == '7':
        # in case of yum, set debuglevel=2 to be sure the output is always
        # same. The format of data is different for various debuglevels
        cmd = ['yum', '--setopt=debuglevel=2']
    else:
        # the verbose mode in dnf always set particular debuglevel, so the
        # output is not affected by the default debug level set on the
        # system
        cmd = ['dnf', '-v']  # On RHEL8 we need to supply an extra switch

    pkg_manager_output = run(cmd, split=True, checked=False)  # The command will certainly fail (does not matter).

    return _parse_loaded_plugins(pkg_manager_output)



from leapp.actors import Actor
from leapp.libraries.actor import checksystemdservicetasks
from leapp.models import SystemdServicesTasks
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, TargetTransactionChecksPhaseTag


class CheckSystemdServicesTasks(Actor):
    """
    Inhibit the upgrade if SystemdServicesTasks tasks are in conflict

    SystemdServicesTasks messages with conflicting requested service states
    could be produced. For example a service could be requested to be both
    - enabled and disabled. This actor inhibits upgrade in such cases.

    Note: We expect that SystemdServicesTasks could be produced even after the
    TargetTransactionChecksPhase (e.g. during the ApplicationsPhase). The
    purpose of this actor is to report collisions in case we can already detect
    them. In case of conflicts caused by messages produced later we just log
    the collisions and the services will end up disabled.
    """

    name = 'check_systemd_services_tasks'
    consumes = (SystemdServicesTasks,)
    produces = (Report,)
    tags = (TargetTransactionChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checksystemdservicetasks.check_conflicts()



from leapp.actors import Actor
from leapp.libraries.actor import updates
from leapp.libraries.common import rpms
from leapp.models import BindFacts, DistributionSignedRPM
from leapp.tags import IPUWorkflowTag, PreparationPhaseTag


class BindUpdate(Actor):
    """
    Actor parsing facts found in configuration and modifying configuration.
    """

    name = 'bind_update'
    consumes = (DistributionSignedRPM, BindFacts)
    produces = ()
    tags = (PreparationPhaseTag, IPUWorkflowTag)

    pkg_names = {'bind', 'bind-sdb', 'bind-pkcs11'}

    def has_bind_package(self):
        """Test any bind server package is installed."""
        for pkg in self.pkg_names:
            if rpms.has_package(DistributionSignedRPM, pkg):
                return True
        return False

    def process(self):
        if not self.has_bind_package():
            self.log.debug('bind is not installed')
            return

        for bindfacts in self.consume(BindFacts):
            updates.update_facts(bindfacts)
            self.log.info('BIND configuration files modified: %s',
                          ', '.join(bindfacts.modified_files))



from leapp.models import InstalledKdeAppsFacts, InstalledRPM, RPM
from leapp.snactor.fixture import current_actor_context

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'

#  KDE apps (only name matters, other values are irrelevant)
okular_RPM = RPM(name='okular', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                 pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')
kdenetwork_RPM = RPM(name='kdenetwork', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER,
                     arch='noarch', pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')
kate_RPM = RPM(name='kate', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
               pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')

# Some other apps to check false detection (only name matters, other values are irrelevant)
epiphany_PRM = RPM(name='epiphany', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                   pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')
polari_RPM = RPM(name='polari', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                 pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')


def test_no_app_present(current_actor_context):
    current_actor_context.feed(InstalledRPM(items=[]))
    current_actor_context.run()
    message = current_actor_context.consume(InstalledKdeAppsFacts)[0]
    assert not message.installed_apps


def test_no_KDE_app_present(current_actor_context):
    current_actor_context.feed(InstalledRPM(items=[epiphany_PRM, polari_RPM]))
    current_actor_context.run()
    message = current_actor_context.consume(InstalledKdeAppsFacts)[0]
    assert not message.installed_apps


def test_only_KDE_apps_present(current_actor_context):
    current_actor_context.feed(InstalledRPM(items=[okular_RPM, kdenetwork_RPM, kate_RPM]))
    current_actor_context.run()
    message = current_actor_context.consume(InstalledKdeAppsFacts)[0]
    assert len(message.installed_apps) == 3


def test_many_apps_present(current_actor_context):
    current_actor_context.feed(InstalledRPM(items=[okular_RPM, kdenetwork_RPM, kate_RPM, epiphany_PRM, polari_RPM]))
    current_actor_context.run()
    message = current_actor_context.consume(InstalledKdeAppsFacts)[0]
    assert len(message.installed_apps) == 3



import mock

from leapp.actors import Actor
from leapp.models import InstalledRPM, Report, RPM


def test_x32_x64(current_actor_context):
    problem_rpms = [
       RPM(name='brlapi', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER", arch='i686',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
       RPM(name='gnome-online-accounts-devel', version='0.1', release='1.sm01', epoch='1',
           packager="RH_PACKAGER", arch='i686', pgpsig='SOME_OTHER_SIG_X'),
       RPM(name='geocode-glib-devel', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER",
           arch='i686', pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186'),
       RPM(name='brlapi', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER", arch='x86_64',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
       RPM(name='gnome-online-accounts-devel', version='0.1', release='1.sm01', epoch='1',
           packager="RH_PACKAGER", arch='x86_64', pgpsig='SOME_OTHER_SIG_X'),
       RPM(name='geocode-glib-devel', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER",
           arch='x86_64',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186')
       ]

    current_actor_context.feed(InstalledRPM(items=problem_rpms))
    current_actor_context.run()
    report = current_actor_context.consume(Report)[0].report
    assert report['title'] == ('Some packages have both 32bit and 64bit version installed which are known to'
                               ' cause rpm transaction test to fail')
    assert {p['title'] for p in report['detail']['related_resources'] if p['scheme'] == 'package'} == \
           {'brlapi.i686', 'gnome-online-accounts-devel.i686', 'geocode-glib-devel.i686'}


def test_1_package(current_actor_context):
    pkg = 'geocode-glib-devel'
    problem_rpms = [
       RPM(name=pkg, version='0.1', release='1.sm01', epoch='1',
           packager="RH_PACKAGER", arch='x86_64', pgpsig='SOME_OTHER_SIG_X'),
       RPM(name=pkg, version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER",
           arch='i686',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186')
    ]
    current_actor_context.feed(InstalledRPM(items=problem_rpms))
    current_actor_context.run()
    report = current_actor_context.consume(Report)[0].report
    assert report['title'] == ('Some packages have both 32bit and 64bit version installed which are known to'
                               ' cause rpm transaction test to fail')
    assert {p['title'] for p in report['detail']['related_resources'] if p['scheme'] == 'package'} == \
           {'{}.i686'.format(pkg)}


def test_x64_only(current_actor_context):
    ok_rpms = [
       RPM(name='brlapi', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER", arch='x86_64',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
       RPM(name='gnome-online-accounts-devel', version='0.1', release='1.sm01', epoch='1',
           packager="RH_PACKAGER", arch='x86_64', pgpsig='SOME_OTHER_SIG_X'),
       RPM(name='geocode-glib-devel', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER",
           arch='x86_64',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186')
       ]

    current_actor_context.feed(InstalledRPM(items=ok_rpms))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_x32_only(current_actor_context):
    ok_rpms = [
       RPM(name='brlapi', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER", arch='i686',
           pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
       RPM(name='gnome-online-accounts-devel', version='0.1', release='1.sm01', epoch='1',
           packager="RH_PACKAGER", arch='i686', pgpsig='SOME_OTHER_SIG_X'),
       RPM(name='geocode-glib-devel', version='0.1', release='1.sm01', epoch='1', packager="RH_PACKAGER",
           arch='i686', pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 5326810137017186'),
       ]

    current_actor_context.feed(InstalledRPM(items=ok_rpms))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



import sys

import pytest

from leapp.libraries.actor import rpmscanner
from leapp.libraries.common import module as module_lib
from leapp.libraries.common import rpms, testutils
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM, RPM
from leapp.snactor.fixture import current_actor_context

no_yum = False
try:
    import yum
except ImportError:
    no_yum = True

no_dnf = False
try:
    import dnf
except ImportError:
    no_dnf = True


# real module streams taken from Fedora 31
ARTIFACTS_AFTERBURN = [
    'afterburn-0:4.2.0-1.module_f31+6825+8330d585.x86_64',
    'afterburn-debuginfo-0:4.2.0-1.module_f31+6825+8330d585.x86_64',
    'rust-afterburn-0:4.2.0-1.module_f31+6825+8330d585.src',
    'rust-afterburn-debugsource-0:4.2.0-1.module_f31+6825+8330d585.x86_64'
]
ARTIFACTS_SUBVERSION_110 = [
    'mod_dav_svn-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'mod_dav_svn-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'python2-subversion-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'python2-subversion-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-0:1.10.6-1.module_f31+5204+aeb0fc0d.src',
    'subversion-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-debugsource-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-devel-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-devel-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-gnome-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-gnome-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-javahl-0:1.10.6-1.module_f31+5204+aeb0fc0d.noarch',
    'subversion-kde-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-kde-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-libs-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-libs-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-perl-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-perl-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-tools-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64',
    'subversion-tools-debuginfo-0:1.10.6-1.module_f31+5204+aeb0fc0d.x86_64'
]
ARTIFACTS_SUBVERSION_113 = [
    'mod_dav_svn-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'mod_dav_svn-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'python2-subversion-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'python2-subversion-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-0:1.13.0-1.module_f31+6955+7c448939.src',
    'subversion-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-debugsource-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-devel-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-devel-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-gnome-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-gnome-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-javahl-0:1.13.0-1.module_f31+6955+7c448939.noarch',
    'subversion-kde-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-kde-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-libs-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-libs-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-perl-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-perl-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-tools-0:1.13.0-1.module_f31+6955+7c448939.x86_64',
    'subversion-tools-debuginfo-0:1.13.0-1.module_f31+6955+7c448939.x86_64'
]


class ModuleMocked(object):
    def __init__(self, name, stream, artifacts):
        self.name = name
        self.stream = stream
        self.artifacts = artifacts

    def getName(self):
        return self.name

    def getStream(self):
        return self.stream

    def getArtifacts(self):
        return self.artifacts


MODULES = [
    ModuleMocked('afterburn', 'rolling', ARTIFACTS_AFTERBURN),
    ModuleMocked('subversion', '1.10', ARTIFACTS_SUBVERSION_110),
    ModuleMocked('subversion', '1.13', ARTIFACTS_SUBVERSION_113)
]


@pytest.mark.skipif(no_yum and no_dnf, reason='yum/dnf is unavailable')
def test_actor_execution(monkeypatch, current_actor_context):
    monkeypatch.setattr(rpmscanner.module_lib, 'get_modules', lambda: [])
    current_actor_context.run()
    assert current_actor_context.consume(InstalledRPM)
    assert current_actor_context.consume(InstalledRPM)[0].items


def test_map_modular_rpms_to_modules_empty(monkeypatch):
    monkeypatch.setattr(module_lib, 'get_modules', lambda: [])
    mapping = rpmscanner.map_modular_rpms_to_modules()
    assert not mapping


def test_map_modular_rpms_to_modules(monkeypatch):
    monkeypatch.setattr(module_lib, 'get_modules', lambda: MODULES)
    mapping = rpmscanner.map_modular_rpms_to_modules()
    assert mapping[
        ('afterburn', '0', '4.2.0', '1.module_f31+6825+8330d585', 'x86_64')
    ] == ('afterburn', 'rolling')
    assert mapping[
        ('subversion', '0', '1.10.6', '1.module_f31+5204+aeb0fc0d', 'x86_64')
    ] == ('subversion', '1.10')
    assert mapping[
        ('subversion', '0', '1.13.0', '1.module_f31+6955+7c448939', 'x86_64')
    ] == ('subversion', '1.13')
    assert not mapping.get(('subversion', '0', '1.13.0', '1.module_f31+6955+7c448939', 'noarch'))
    assert not mapping.get(('subversion', '0', '1.13.1', '1.module_f31+6955+7c448939', 'x86_64'))
    assert not mapping.get(('subversion', '1', '1.13.0', '1.module_f31+6955+7c448939', 'x86_64'))


INSTALLED_RPMS = [
    ('afterburn|4.2.0|1.module_f31+6825+8330d585|0|Fedora Project|x86_64|'
     'RSA/SHA256, Wed 16 Oct 2019 12:49:08 AM CEST, Key ID 50cb390b3c3359c4'),
    ('subversion|1.10.6|1.module_f31+5204+aeb0fc0d|0|Fedora Project|x86_64|'
     'RSA/SHA256, Thu 25 Jul 2019 01:41:52 PM CEST, Key ID 50cb390b3c3359c4'),
    # non-modular, epoch
    ('tcpdump|4.9.3|2.fc31|14|Fedora Project|x86_64|'
     'RSA/SHA256, Wed 22 Jul 2020 12:25:15 PM CEST, Key ID 50cb390b3c3359c4'),
    # non-modular, no epoch
    ('passwd|0.80|7.fc31|0|Fedora Project|x86_64|'
     'RSA/SHA256, Wed 04 Dec 2019 08:48:43 PM CET, Key ID 50cb390b3c3359c4')
]


PACKAGE_REPOS = {
    'afterburn': 'repo1',
    'subversion': 'repo2',
    'tcpdump': 'repo2'
}


def test_process(monkeypatch):
    monkeypatch.setattr(module_lib, 'get_modules', lambda: MODULES)
    monkeypatch.setattr(rpmscanner, 'get_package_repository_data', lambda: PACKAGE_REPOS)
    monkeypatch.setattr(rpms, 'get_installed_rpms', lambda: INSTALLED_RPMS)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())

    rpmscanner.process()
    assert api.produce.called
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], InstalledRPM)
    items = {i.name: i for i in api.produce.model_instances[0].items}
    assert len(items) == 4

    assert items['afterburn'].epoch == '0'
    assert items['afterburn'].version == '4.2.0'
    assert items['afterburn'].release == '1.module_f31+6825+8330d585'
    assert items['afterburn'].arch == 'x86_64'
    assert items['afterburn'].module == 'afterburn'
    assert items['afterburn'].stream == 'rolling'

    assert items['subversion'].epoch == '0'
    assert items['subversion'].version == '1.10.6'
    assert items['subversion'].release == '1.module_f31+5204+aeb0fc0d'
    assert items['subversion'].arch == 'x86_64'
    assert items['subversion'].module == 'subversion'
    assert items['subversion'].stream == '1.10'

    assert items['tcpdump'].epoch == '14'
    assert items['tcpdump'].version == '4.9.3'
    assert items['tcpdump'].release == '2.fc31'
    assert items['tcpdump'].arch == 'x86_64'
    assert not items['tcpdump'].module
    assert not items['tcpdump'].stream

    assert items['passwd'].epoch == '0'
    assert items['passwd'].version == '0.80'
    assert items['passwd'].release == '7.fc31'
    assert items['passwd'].arch == 'x86_64'
    assert not items['passwd'].module
    assert not items['passwd'].stream



import json
import os

from leapp.libraries.common import repofileutils

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def test_invert_dict():
    input_dict = {1: ['a', 'b'], 2: ['b'], 3: []}
    inv_dict = repofileutils._invert_dict(input_dict)
    assert inv_dict == {'a': [1], 'b': [1, 2]}


def test_parse_repofile():
    repofile = repofileutils.parse_repofile(os.path.join(CUR_DIR, 'sample_repos.txt'))

    repo_appstream = [repo for repo in repofile.data if repo.repoid == 'AppStream'][0]
    assert repo_appstream.name == 'CentOS-$releasever - AppStream'
    assert repo_appstream.baseurl is None  # comments shouldn't get parsed
    assert repo_appstream.metalink is None
    assert repo_appstream.mirrorlist == ('http://mirrorlist.centos.org/?release=$releasever'
                                         '&arch=$basearch&repo=AppStream&infra=$infra')
    assert repo_appstream.enabled is True
    additional_appstream = json.loads(repo_appstream.additional_fields)
    assert additional_appstream['gpgcheck'] == '1'
    assert additional_appstream['gpgkey'] == 'file:///etc/pki/rpm-gpg/RPM-GPG-KEY-centosofficial'
    assert additional_appstream['cost'] == '77'
    assert additional_appstream.get('baseurl') is None

    repo_leapp = [repo for repo in repofile.data if repo.repoid == 'leapp-copr'][0]
    assert repo_leapp.name == 'Copr repo for devel Leapp builds'
    assert repo_leapp.baseurl == 'http://coprbe.devel.redhat.com/results/oam-group/leapp/rhel-7-x86_64/'
    assert repo_leapp.metalink is None
    assert repo_leapp.mirrorlist is None
    assert repo_leapp.enabled is False
    additional_leapp = json.loads(repo_leapp.additional_fields)
    assert additional_leapp['type'] == 'rpm-md'
    assert additional_leapp['skip_if_unavailable'] == 'True'
    assert additional_leapp['gpgcheck'] == '0'
    assert additional_leapp['gpgkey'] == 'http://coprbe.devel.redhat.com/results/oam-group/leapp/pubkey.gpg'
    assert additional_leapp['repo_gpgcheck'] == '0'
    assert additional_leapp['enabled_metadata'] == '1'
    assert len(additional_leapp) == 6

    assert len([repo for repo in repofile.data if repo.repoid == 'spe-ci_al.cha:rs']) == 1

    repos_duplicate = [repo for repo in repofile.data if repo.repoid == 'duplicate']
    assert len(repos_duplicate) == 1  # only one instance got through
    assert repos_duplicate[0].name == 'Duplicate 2'  # and it's the latter one



import logging
import os

from leapp.repository.manager import RepositoryManager
from leapp.repository.scan import find_and_scan_repositories
from leapp.utils.repository import find_repository_basedir, get_repository_id

logger = logging.getLogger(__name__)
logging.getLogger("asyncio").setLevel(logging.INFO)
logging.getLogger("parso").setLevel(logging.INFO)


def _load_and_add_repo(manager, repo_path):
    repo = find_and_scan_repositories(
        repo_path,
        include_locals=True
    )
    unloaded = set()
    loaded = {r.repo_id for r in manager.repos}
    if hasattr(repo, 'repos'):
        for repo in repo.repos:
            if not manager.repo_by_id(repo.repo_id):
                manager.add_repo(repo)
                unloaded.add(repo.repo_id)
    else:
        manager.add_repo(repo)
    if not loaded:
        manager.load(skip_actors_discovery=True)
    else:
        for repo_id in unloaded:
            manager.repo_by_id(repo_id).load(skip_actors_discovery=True)


def pytest_collectstart(collector):
    if collector.nodeid:
        current_repo_basedir = find_repository_basedir(str(collector.fspath))
        if not current_repo_basedir:
            # This is not a repository
            return
        if not hasattr(collector.session, "leapp_repository"):
            collector.session.leapp_repository = RepositoryManager()
            collector.session.repo_base_dir = current_repo_basedir
            _load_and_add_repo(collector.session.leapp_repository, current_repo_basedir)
        else:
            if not collector.session.leapp_repository.repo_by_id(
                get_repository_id(current_repo_basedir)
            ):
                _load_and_add_repo(collector.session.leapp_repository, current_repo_basedir)

        # we're forcing the actor context switch only when traversing new
        # actor
        if "/actors/" in str(collector.fspath) and (
            not hasattr(collector.session, "current_actor_path")
            or collector.session.current_actor_path + os.sep
            not in str(collector.fspath)
        ):
            actor = None
            for a in collector.session.leapp_repository.actors:
                if a.full_path == collector.fspath.dirpath().dirname:
                    actor = a
                    break

            if not actor:
                logger.info("No actor found, exiting collection...")
                return
            # we need to tear down the context from the previous
            # actor
            try:
                collector.session.current_actor_context.__exit__(
                    None, None, None
                )
            except AttributeError:
                pass
            else:
                logger.info(
                    "Actor %r context teardown complete",
                    collector.session.current_actor.name,
                )

            logger.info("Injecting actor context for %r", actor.name)
            collector.session.current_actor = actor
            collector.session.current_actor_context = actor.injected_context()
            collector.session.current_actor_context.__enter__()
            collector.session.current_actor_path = (
                collector.session.current_actor.full_path
            )
            logger.info("Actor %r context injected", actor.name)


def pytest_runtestloop(session):
    try:
        session.current_actor_context.__exit__(None, None, None)
        logger.info(
            "Actor %r context teardown complete", session.current_actor.name,
        )
    except AttributeError:
        pass



from os import listdir, path

from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import SapHanaInfo, SapHanaInstanceInfo, SapHanaManifestEntry

HANA_BASE_PATH = '/hana/shared'
HANA_MANIFEST_PATH = 'exe/linuxx86_64/hdb/manifest'
HANA_SAPCONTROL_PATH = 'exe/linuxx86_64/hdb/sapcontrol'


def perform_sap_hana_scan():
    """
    Produces a message with details collected around SAP HANA.
    """

    api.produce(search_sap_hana_instances())


def parse_manifest(path):
    """ Parses a SAP HANA manifest into a dictionary """
    def _decoded(s):
        """
        Compatibility between Python2 and 3 - Python 2 has no str.decode but we can process str directly.
        Python3 needs the byte data to be decoded so we can process the string data.
        """
        if hasattr(s, 'decode'):
            return s.decode('utf-8')
        return s

    data = []
    try:
        with open(path, 'r') as f:
            for line in _decoded(f.read()).split('\n'):
                try:
                    key, value = line.split(':', 1)
                except ValueError:
                    # Most likely an empty line, but we're being permissive here and ignore failures.
                    # In the end it's all about having the right values available.
                    if line:
                        api.current_logger().warning(
                            'Failed to parse line in manifest: {file}. Line was: `{line}`'.format(file=path,
                                                                                                  line=line),
                            exc_info=True)
                    continue
                data.append(SapHanaManifestEntry(key=key, value=value.strip()))
    except OSError:
        return None
    return data


def search_sap_hana_instances():
    """
    Searches for all instances of SAP HANA on the system and gets the information for it

    This code will go through all entries in /hana/shared and checks for all instances within.
    For each instance it will check the status and record the location.

    :return: SapHanaInfo
    """
    # Keeps track of found instances
    result = []
    # Keeps track if any instance is running
    any_running = False
    # Keeps track if any SAP HANA installation was found. In theory it could be also derived from len(result) != 0
    installed = False
    if path.isdir(HANA_BASE_PATH):
        for entry in listdir(HANA_BASE_PATH):

            entry_path = path.join(HANA_BASE_PATH, entry)
            sapcontrol_path = path.join(entry_path, HANA_SAPCONTROL_PATH)
            entry_manifest_path = path.join(entry_path, HANA_MANIFEST_PATH)
            if path.isfile(entry_manifest_path):
                # We found the manifest file in the expected relative path.
                # Now we are going to look for instance directories.
                for instance in listdir(entry_path):
                    instance_number = None
                    if 'HDB' in instance:
                        # We found an instance. Instance directories follow HDB[0-9][0-9] naming pattern,
                        # where the numbers represent the instance number.
                        instance_number = instance[-2:]
                    if not instance_number:
                        # This is not a folder we are interested in
                        continue
                    # Now obviously SAP HANA is installed
                    installed = True
                    # We can derive the admin name from `entry` directory name
                    admin_name = '{}adm'.format(entry.lower())
                    # Retrieving the status of this instance
                    running = get_instance_status(instance_number, sapcontrol_path, admin_name)
                    # Update the variable that tracks all instances if any is running.
                    # This makes the inhibitor code easier later.
                    any_running = any_running or running
                    # Append the found instance to the list
                    result.append(
                        SapHanaInstanceInfo(
                            name=entry,
                            manifest=parse_manifest(entry_manifest_path),
                            path=entry_path,
                            instance_number=instance_number,
                            running=running,
                            admin=admin_name
                        )
                    )
    # Return the results
    return SapHanaInfo(instances=result, running=any_running, installed=installed)


def get_instance_status(instance_number, sapcontrol_path, admin_name):
    """ Gets the status for the instance given """
    try:
        # Executes sapcontrol in the context of the instance admin user to retrieve the process list for the given
        # SAP HANA instance.
        # GetProcessList has some oddities, like returning non zero exit codes with special meanings.
        # Exit code 3 = All processes are running correctly
        # Exit code 4 = All processes stopped
        # Other exit codes aren't handled at this time and it's assumed that SAP HANA is possibly in some unusual
        # state. Such as starting/stopping but also that it is in some kind of failure state.
        output = run([
            'sudo', '-u', admin_name, sapcontrol_path, '-nr', instance_number, '-function', 'GetProcessList'],
            checked=False)
        if output['exit_code'] == 3:
            # GetProcessList succeeded, all processes running correctly
            return True
        if output['exit_code'] == 4:
            # GetProcessList succeeded, all processes stopped
            return False
        # SAP HANA might be somewhere in between (Starting/Stopping)
        # In that case there are always more than 7 lines.
        return len(output['stdout'].split('\n')) > 7
    except CalledProcessError:
        api.current_logger().warning(
            'Failed to retrieve SAP HANA instance status from sapcontrol - Considering it as not running.')
        return False
